#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import json
import unittest

from . import collector as TT


class NoDataCollector(TT.DataCollector):
    kind = 'not_exist'


class TestCase(unittest.TestCase):

    def test_load_datasets_failures(self):
        collector = NoDataCollector()
        with self.assertRaises(ValueError):
            collector.init()

    def test_load_datasets(self):
        collector = TT.DataCollector()
        collector.init()
        res = collector.datasets

        inp_refs = list(collector.root.glob(f'*/{collector.pattern}'))
        assert bool(inp_refs)

        e_refs = list(collector.root.glob('*/e/*.json'))
        assert bool(e_refs)

        o_refs = list(collector.root.glob('*/o/*.json'))
        assert bool(o_refs)

        s_refs = list(collector.root.glob('*/s/*.json'))
        assert bool(s_refs)

        for inp in inp_refs:
            self.assertTrue(any(inp in td.inputs for td in res))

        for e_file in e_refs:
            e_ref = json.load(e_file.open())
            self.assertTrue(any(e_ref == td.exp for td in res))

        for o_file in o_refs:
            o_ref = json.load(o_file.open())
            self.assertTrue(any(o_ref == td.opts for td in res))

        for s_file in s_refs:
            self.assertTrue(any(s_file == td.scm for td in res))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-import,import-error,invalid-name
r"""Public APIs to dump configurations data.
"""
import typing

from ..backend import base
from ..common import InDataT


MaybeDataT = typing.Optional[InDataT]
ParserT = base.Parser

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods
import os
import tests.backend.common as TBC
try:
    import anyconfig.backend.yaml.pyyaml as TT
except ImportError:
    import unittest
    raise unittest.SkipTest

from .common import CNF_S, CNF


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf = CNF
    cnf_s = CNF_S

    opts = dict(typ="rt", pure=True,
                preserve_quotes=True,
                indent=dict(mapping=4, sequence=4, offset=2))

    setattr(psr, "_dict_opts", opts)


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):  # noqa: N801

    load_options = dict(ac_safe=True, Loader=TT.yaml.loader.Loader)
    dump_options = dict(ac_safe=True)
    empty_patterns = [('', {}), (' ', {}), ('[]', []),
                      ("#%s#%s" % (os.linesep, os.linesep), {})]


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):  # noqa: N801

    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
"""Test cases for anyconfig.parser.parse_attrlist.
"""
import anyconfig.parser as TT

from . import common


class TestCase(common.TestCase):
    kind = 'attrlist'
    pattern = '*.txt'

    def test_parse_attrlist(self):
        for data in self.each_data():
            self.assertEqual(
                TT.parse_attrlist(data.inp, **data.opts),
                data.exp,
                data
            )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
"""File based test data collector - utility functions.
"""
import pathlib
import typing
import warnings

from ... import base
#    load_data, maybe_data_path,

from .datatypes import TData


# .. seealso:: tests.api.multi_load.datatypes
MaybeDataT = typing.Optional[
    typing.Union[str, pathlib.Path, typing.Dict[str, typing.Any]]
]


def load_data_or_path(datadir: pathlib.Path,
                      should_exist: typing.Iterable[str] = (),
                      load: bool = True,
                      default: typing.Optional[typing.Any] = None
                      ) -> MaybeDataT:
    """
    Load data from a file in the ``datadir`` of which name matches ``pattern``.
    """
    maybe_file = base.maybe_data_path(datadir, '*', should_exist=should_exist)
    if maybe_file is None:
        return default

    if load:
        return base.load_data(maybe_file)

    return maybe_file


def each_data_from_dir(datadir: pathlib.Path,
                       pattern: str = '*.json',
                       should_exist: typing.Iterable[str] = ()
                       ) -> typing.Iterator[TData]:
    """
    Yield a collection of paths of data files under given dir.
    """
    if not datadir.is_dir():
        raise ValueError(f'Not look a data dir: {datadir!s}')

    for subdir in sorted(datadir.glob('*')):
        if not subdir.is_dir():
            warnings.warn(f'Not looks a dir: {subdir!s}')
            continue

        if not bool(list(subdir.glob('*.*'))):
            warnings.warn(f'No data in subdir: {subdir!s}')
            continue

        yield TData(
            subdir,
            sorted(
                inp for inp in subdir.glob(pattern) if inp.is_file()
            ),
            load_data_or_path(subdir / 'e', should_exist, default={}),
            load_data_or_path(subdir / 'o', should_exist, default={}),
            load_data_or_path(subdir / 's', should_exist, load=False),
            load_data_or_path(subdir / 'q', should_exist, default=''),
            load_data_or_path(subdir / 'c', should_exist, default={}),
        )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""misc global constants, variables, classes and so on.
"""
try:
    from .jinja2 import try_render
    SUPPORTED: bool = True
except ImportError:  # jinja2 may not be available.
    SUPPORTED = False  # type: ignore

    def try_render(*_args, **_kwargs) -> None:  # type: ignore
        """Dummy function does nothing but returns None.
        """
        return None


__all__ = [
    'try_render',
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2015 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
import pathlib
import subprocess
import tempfile
import unittest


SCRIPT_TO_USE_ANYCONFIG = """\
#! /usr/bin/env python
import anyconfig

c = anyconfig.load("/") or {}
anyconfig.dump(c, "/dev/null", "yaml")
"""


def check_output(cmd):
    devnull = open('/dev/null', 'w')
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=devnull)
    return proc.communicate()[0]


class TestCase(unittest.TestCase):

    def test_00_run_script(self):
        with tempfile.TemporaryDirectory(prefix='anyconfig-tests-') as tmpdir:
            script = pathlib.Path(tmpdir) / "a.py"
            script.write_text(SCRIPT_TO_USE_ANYCONFIG)

            out = check_output(["python", str(script)])
            self.assertTrue(out in (b'', ''))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Abstract implementation of backend modules:

Backend module must implement a parser class inherits :class:`Parser` or its
children classes of this module and override all or some of the methods as
needed:

  - :meth:`load_from_string`: Load config from string
  - :meth:`load_from_stream`: Load config from a file or file-like object
  - :meth:`load_from_path`: Load config from file of given path
  - :meth:`dump_to_string`: Dump config as a string
  - :meth:`dump_to_stream`: Dump config to a file or file-like object
  - :meth:`dump_to_path`: Dump config to a file of given path
"""
import typing

from ...models import processor
from ...utils import is_dict_like
from .datatypes import (
    InDataExT, GenContainerT
)
from .dumpers import (
    DumperMixin, ToStringDumperMixin, ToStreamDumperMixin
)
from .loaders import (
    LoaderMixin, FromStringLoaderMixin, FromStreamLoaderMixin
)


class Parser(LoaderMixin, DumperMixin, processor.Processor):
    """
    Abstract parser to provide basic implementation of some methods, interfaces
    and members.

    - _type: Parser type indicate which format it supports
    - _priority: Priority to select it if there are other parsers of same type
    - _extensions: File extensions of formats it supports
    - _open_flags: Opening flags to read and write files

    .. seealso:: the doc of :class:`anyconfig.models.processor.Processor`
    """
    _cid: str = 'base'


class StringParser(Parser, FromStringLoaderMixin, ToStringDumperMixin):
    """
    Abstract parser based on :meth:`load_from_string` and
    :meth:`dump_to_string`.

    Parser classes inherit this class must define these methods.
    """


class StreamParser(Parser, FromStreamLoaderMixin, ToStreamDumperMixin):
    """
    Abstract parser based on :meth:`load_from_stream` and
    :meth:`dump_to_stream`.

    Parser classes inherit this class must define these methods.
    """


LoadFnT = typing.Callable[..., InDataExT]
DumpFnT = typing.Callable[..., typing.Optional[str]]


def load_with_fn(load_fn: typing.Optional[LoadFnT],
                 content_or_strm: typing.Union[str, typing.IO],
                 container: GenContainerT,
                 allow_primitives: bool = False,
                 **options) -> InDataExT:
    """
    Load data from given string or stream 'content_or_strm'.

    :param load_fn: Callable to load data
    :param content_or_strm: data content or stream provides it
    :param container: callble to make a container object
    :param allow_primitives:
        True if the parser.load* may return objects of primitive data types
        other than mapping types such like JSON parser
    :param options: keyword options passed to 'load_fn'

    :return: container object holding data
    """
    if load_fn is None:
        raise TypeError('The first argument "load_fn" must be a callable!')

    ret = load_fn(content_or_strm, **options)
    if is_dict_like(ret):
        return container() if (ret is None or not ret) else container(ret)

    return ret if allow_primitives else container(ret)


def dump_with_fn(dump_fn: typing.Optional[DumpFnT],
                 data: InDataExT, stream: typing.Optional[typing.IO],
                 **options) -> typing.Optional[str]:
    """
    Dump 'data' to a string if 'stream' is None, or dump 'data' to a file or
    file-like object 'stream'.

    :param dump_fn: Callable to dump data
    :param data: Data to dump
    :param stream:  File or file like object or None
    :param options: optional keyword parameters

    :return: String represents data if stream is None or None
    """
    if dump_fn is None:
        raise TypeError('The first argument "dump_fn" must be a callable!')

    if stream is None:
        return dump_fn(data, **options)

    return dump_fn(data, stream, **options)


class StringStreamFnParser(Parser, FromStreamLoaderMixin, ToStreamDumperMixin):
    """
    Abstract parser utilizes load and dump functions each backend module
    provides such like json.load{,s} and json.dump{,s} in JSON backend.

    Parser classes inherit this class must define the followings.

    - _load_from_string_fn: Callable to load data from string
    - _load_from_stream_fn: Callable to load data from stream (file object)
    - _dump_to_string_fn: Callable to dump data to string
    - _dump_to_stream_fn: Callable to dump data to stream (file object)

    .. note::
       Callables have to be wrapped with :func:`to_method` to make 'self'
       passed to the methods created from them ignoring it.

    :seealso: :class:`anyconfig.backend.json.Parser`
    """
    _load_from_string_fn: typing.Optional[LoadFnT] = None
    _load_from_stream_fn: typing.Optional[LoadFnT] = None
    _dump_to_string_fn: typing.Optional[DumpFnT] = None
    _dump_to_stream_fn: typing.Optional[DumpFnT] = None

    def load_from_string(self, content: str, container: GenContainerT,
                         **options) -> InDataExT:
        """
        Load configuration data from given string 'content'.

        :param content: Configuration string
        :param container: callble to make a container object
        :param options: keyword options passed to '_load_from_string_fn'

        :return: container object holding the configuration data
        """
        return load_with_fn(self._load_from_string_fn, content, container,
                            allow_primitives=self.allow_primitives(),
                            **options)

    def load_from_stream(self, stream: typing.IO, container: GenContainerT,
                         **options) -> InDataExT:
        """
        Load data from given stream 'stream'.

        :param stream: Stream provides configuration data
        :param container: callble to make a container object
        :param options: keyword options passed to '_load_from_stream_fn'

        :return: container object holding the configuration data
        """
        return load_with_fn(self._load_from_stream_fn, stream, container,
                            allow_primitives=self.allow_primitives(),
                            **options)

    def dump_to_string(self, cnf: InDataExT, **kwargs) -> str:
        """
        Dump config 'cnf' to a string.

        :param cnf: Configuration data to dump
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: string represents the configuration
        """
        return dump_with_fn(self._dump_to_string_fn, cnf, None,
                            **kwargs)  # type: ignore

    def dump_to_stream(self, cnf: InDataExT, stream: typing.IO,
                       **kwargs) -> None:
        """
        Dump config 'cnf' to a file-like object 'stream'.

        TODO: How to process socket objects same as file objects ?

        :param cnf: Configuration data to dump
        :param stream:  Config file or file like object
        :param kwargs: optional keyword parameters to be sanitized :: dict
        """
        dump_with_fn(self._dump_to_stream_fn, cnf, stream,
                     **kwargs)  # type: ignore

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""misc global constants, variables, classes and so on.
"""
try:
    from .jsonschema import validate, is_valid, gen_schema
    SUPPORTED: bool = True
except ImportError:
    from .default import validate, is_valid, gen_schema
    SUPPORTED = False  # type: ignore


__all__ = [
    'validate', 'is_valid', 'gen_schema', 'SUPPORTED'
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
"""Misc parsers
"""
import re
import typing
import warnings


INT_PATTERN: typing.Pattern = re.compile(r"^(\d|([1-9]\d+))$")
FLOAT_PATTERN: typing.Pattern = re.compile(r"^\d+[\.]\d+$")
BOOL_PATTERN: typing.Pattern = re.compile(r"^(true|false)$", re.I)
STR_PATTERN: typing.Pattern = re.compile(r"^['\"](.*)['\"]$")


PrimitiveT = typing.Union[str, int, float, bool]
PrimitivesT = typing.List[PrimitiveT]


def parse_single(str_: typing.Optional[str]) -> PrimitiveT:
    """
    Very simple parser to parse an expression represents a primitive value.

    >>> parse_single(None)
    ''
    >>> parse_single('0')
    0
    >>> parse_single('123')
    123
    >>> parse_single('True')
    True
    >>> parse_single('a string')
    'a string'
    >>> parse_single('"a string"')
    'a string'
    >>> parse_single("'a string'")
    'a string'
    >>> parse_single('0.1')
    0.1
    >>> parse_single('    a string contains extra whitespaces     ')
    'a string contains extra whitespaces'
    """
    if str_ is None:
        return ''

    str_ = str_.strip()

    if not str_:
        return ''

    if BOOL_PATTERN.match(str_) is not None:
        return bool(str_)

    if INT_PATTERN.match(str_) is not None:
        return int(str_)

    if FLOAT_PATTERN.match(str_) is not None:
        return float(str_)

    if STR_PATTERN.match(str_) is not None:
        return str_[1:-1]

    return str_


def parse_list(str_: str, sep: str = ',') -> PrimitivesT:
    """
    Simple parser to parse an expression ``str_`` contains a list of str-es
    separated with ``sep``, represents a list of primitive values.

    >>> parse_list('')
    []
    >>> parse_list('1')
    [1]
    >>> parse_list('a,b')
    ['a', 'b']
    >>> parse_list('1,2')
    [1, 2]
    >>> parse_list('a,b,')
    ['a', 'b']
    """
    return [parse_single(x) for x in str_.split(sep) if x]


AttrValsT = typing.Tuple[str, typing.Union[PrimitivesT, PrimitiveT]]


def attr_val_itr(str_: str, avs_sep: str = ':', vs_sep: str = ',',
                 as_sep: str = ';') -> typing.Iterator[AttrValsT]:
    """
    Atrribute and value pair parser.

    :param str_: String represents a list of pairs of attribute and value
    :param avs_sep: char to separate attribute and values
    :param vs_sep: char to separate values
    :param as_sep: char to separate attributes
    """
    for rel in parse_list(str_, as_sep):
        rel = typing.cast(str, rel)
        if avs_sep not in rel or rel.endswith(avs_sep):
            continue

        (_attr, _values, *_rest) = parse_list(rel, avs_sep)

        if _rest:
            warnings.warn(f'Extra strings {_rest!s} in {rel!s}'
                          f'It should be in the form of attr{avs_sep}value.')

        _attr = typing.cast(str, _attr)

        if vs_sep in str(_values):
            yield (_attr, parse_list(typing.cast(str, _values), vs_sep))
        elif _values:
            yield (_attr, typing.cast(PrimitiveT, _values))


def parse_attrlist_0(str_: str, avs_sep: str = ':', vs_sep: str = ',',
                     as_sep: str = ';') -> typing.List[AttrValsT]:
    """
    Simple parser to parse expressions in the form of
    [ATTR1:VAL0,VAL1,...;ATTR2:VAL0,VAL2,..].

    :param str_: input string
    :param avs_sep:  char to separate attribute and values
    :param vs_sep:  char to separate values
    :param as_sep:  char to separate attributes

    :return:
        a list of tuples of (key, value | [value])
            where key = (Int | String | ...),
            value = (Int | Bool | String | ...) | [Int | Bool | String | ...]

    >>> parse_attrlist_0('a:1')
    [('a', 1)]
    >>> parse_attrlist_0('a:1;b:xyz')
    [('a', 1), ('b', 'xyz')]
    >>> parse_attrlist_0('requires:bash,zsh')
    [('requires', ['bash', 'zsh'])]
    >>> parse_attrlist_0('obsoletes:sysdata;conflicts:sysdata-old')
    [('obsoletes', 'sysdata'), ('conflicts', 'sysdata-old')]
    """
    return list(attr_val_itr(str_, avs_sep, vs_sep, as_sep))


AttrValsDictT = typing.Dict[str, typing.Union[PrimitivesT, PrimitiveT]]


def parse_attrlist(str_: str, avs_sep: str = ':', vs_sep: str = ',',
                   as_sep: str = ';') -> AttrValsDictT:
    """
    Simple parser to parse expressions in the form of
    [ATTR1:VAL0,VAL1,...;ATTR2:VAL0,VAL2,..].

    :param str_: input string
    :param avs_sep:  char to separate attribute and values
    :param vs_sep:  char to separate values
    :param as_sep:  char to separate attributes

    >>> parse_attrlist('requires:bash,zsh')
    {'requires': ['bash', 'zsh']}
    """
    return dict(parse_attrlist_0(str_, avs_sep, vs_sep, as_sep))


ResultsT = typing.Union[
    PrimitiveT,
    PrimitivesT,
    AttrValsDictT
]


def parse(str_: typing.Optional[str],
          lsep: str = ',', avsep: str = ':', vssep: str = ',',
          avssep: str = ';') -> ResultsT:
    """Generic parser"""
    if str_ is None or not str_:
        return parse_single(str_)

    if avsep in str_:
        return parse_attrlist(str_, avsep, vssep, avssep)
    if lsep in str_:
        return parse_list(str_, lsep)

    return parse_single(str_)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Test cases for anyconfig.utils.files.
"""
import pathlib
import unittest

import anyconfig.utils.files as TT


class TestCase(unittest.TestCase):

    def test_is_io_stream(self):
        ies = (
            (open(__file__), True),
            (__file__, False),
            ([__file__], False),
            (pathlib.Path(__file__), False),
            ([pathlib.Path(__file__)], False),
        )
        for inp, exp in ies:
            res = TT.is_io_stream(inp)
            (self.assertTrue if exp else self.assertFalse)(res)

    def test_get_path_from_stream(self):
        this = __file__

        with pathlib.Path(this).open() as strm:
            self.assertEqual(TT.get_path_from_stream(strm), this)

        with self.assertRaises(ValueError):
            TT.get_path_from_stream(this)

        self.assertEqual(TT.get_path_from_stream(this, safe=True), '')

# vim:sw=4:ts=4:et:



#
# Jinja2 (http://jinja.pocoo.org) based template renderer.
#
# Copyright (C) Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=wrong-import-position,wrong-import-order
"""anyconfig.template.jinja2_ module

Template rendering module for jinja2-based template config files.
"""
import pathlib
import os
import typing
import warnings

import jinja2
import jinja2.exceptions

from .. import utils


# .. seealso:: jinja2.loaders.FileSystemLoader.__init__
PathsT = typing.List[typing.Union[str, pathlib.Path]]
MaybePathsT = typing.Optional[PathsT]
MaybeContextT = typing.Optional[typing.Dict[str, typing.Any]]
MaybeFiltersT = typing.Optional[typing.Iterable[typing.Callable]]

RENDER_S_OPTS: typing.List[str] = ['ctx', 'paths', 'filters']
RENDER_OPTS = RENDER_S_OPTS + ['ask']


def tmpl_env(paths: MaybePathsT) -> jinja2.Environment:
    """
    :param paths: A list of template search paths
    """
    if paths is None:
        paths = []

    return jinja2.Environment(
        loader=jinja2.FileSystemLoader([str(p) for p in paths])
    )


def make_template_paths(template_file: pathlib.Path,
                        paths: MaybePathsT = None
                        ) -> typing.List[str]:
    """
    Make up a list of template search paths from given 'template_file'
    (absolute or relative path to the template file) and/or 'paths' (a list of
    template search paths given by user).

    NOTE: User-given 'paths' will take higher priority over a dir of
    template_file.

    :param template_file: Absolute or relative path to the template file
    :param paths: A list of template search paths
    :return: List of template paths ([str])

    >>> p = pathlib.Path
    >>> make_template_paths(p('/path/to/a/template'))
    ['/path/to/a']
    >>> make_template_paths(p('/path/to/a/template'), ['/tmp'])
    ['/path/to/a', '/tmp']
    >>> os.chdir('/tmp')
    >>> make_template_paths(p('./path/to/a/template'))
    ['/tmp/path/to/a']
    >>> make_template_paths(p('./path/to/a/template'), ['/tmp'])
    ['/tmp/path/to/a', '/tmp']
    """
    tmpldir = str(template_file.parent.resolve())
    if paths:
        return [tmpldir] + [str(p) for p in paths]

    return [tmpldir]


def render_s(tmpl_s: str, ctx: MaybeContextT = None,
             paths: MaybePathsT = None,
             filters: MaybeFiltersT = None
             ) -> str:
    """
    Compile and render given template string 'tmpl_s' with context 'context'.

    :param tmpl_s: Template string
    :param ctx: Context dict needed to instantiate templates
    :param paths: Template search paths
    :param filters: Custom filters to add into template engine
    :return: Compiled result (str)

    >>> render_s('aaa') == 'aaa'
    True
    >>> s = render_s('a = {{ a }}, b = "{{ b }}"', {'a': 1, 'b': 'bbb'})
    >>> assert s == 'a = 1, b = "bbb"'
    """
    if paths is None:
        paths = [os.curdir]

    # .. seealso:: jinja2.environment._environment_sanity_check
    try:
        env = tmpl_env(paths)
    except AssertionError as exc:
        warnings.warn(
            f'Something went wrong with: paths={paths!r}, exc={exc!s}'
        )
        return tmpl_s

    if filters is not None:
        env.filters.update(filters)

    if ctx is None:
        ctx = {}

    return typing.cast(jinja2.Environment, tmpl_env(paths)
                       ).from_string(tmpl_s).render(**ctx)


def render_impl(template_file: pathlib.Path, ctx: MaybeContextT = None,
                paths: MaybePathsT = None, filters: MaybeFiltersT = None
                ) -> str:
    """
    :param template_file: Absolute or relative path to the template file
    :param ctx: Context dict needed to instantiate templates
    :param filters: Custom filters to add into template engine
    :return: Compiled result (str)
    """
    env = tmpl_env(make_template_paths(template_file, paths))  # type: ignore

    if env is None:
        return open(template_file).read()

    if filters is not None:
        env.filters.update(filters)

    if ctx is None:
        ctx = {}

    return env.get_template(pathlib.Path(template_file).name).render(**ctx)


def render(filepath: str, ctx: MaybeContextT = None,
           paths: MaybePathsT = None,
           ask: bool = False,
           filters: MaybeFiltersT = None) -> str:
    """
    Compile and render template and return the result as a string.

    :param template_file: Absolute or relative path to the template file
    :param ctx: Context dict needed to instantiate templates
    :param paths: Template search paths
    :param ask: Ask user for missing template location if True
    :param filters: Custom filters to add into template engine
    :return: Compiled result (str)
    """
    fpath = pathlib.Path(filepath)
    try:
        return render_impl(fpath, ctx, paths, filters)
    except jinja2.exceptions.TemplateNotFound as mtmpl:
        if not ask:
            raise

        usr_tmpl = input(
            f"{os.linesep}*** Missing template '{mtmpl}'. Please enter "
            "absolute or relative path starts from '.' to the template file: "
        )
        usr_tmpl_2 = pathlib.Path(usr_tmpl.strip()).resolve()
        paths_2 = make_template_paths(usr_tmpl_2, paths)

        return render_impl(usr_tmpl_2, ctx, paths_2, filters)  # type: ignore


def try_render(filepath: typing.Optional[str] = None,
               content: typing.Optional[str] = None,
               **options) -> typing.Optional[str]:
    """
    Compile and render template and return the result as a string.

    :param filepath: Absolute or relative path to the template file
    :param content: Template content (str)
    :param options: Keyword options passed to :func:`render` defined above.
    :return: Compiled result (str) or None
    """
    if filepath is None and content is None:
        raise ValueError("Either 'path' or 'content' must be some value!")

    try:
        if content is None:
            render_opts = utils.filter_options(RENDER_OPTS, options)
            return render(typing.cast(str, filepath), **render_opts)

        render_s_opts = utils.filter_options(RENDER_S_OPTS, options)
        return render_s(content, **render_s_opts)

    except Exception as exc:  # pylint: disable=broad-except
        if filepath:
            tmpl_s = filepath
        else:
            tmpl_s = typing.cast(str, content)[:10] + ' ...'

        warnings.warn(
            f"Failed to compile '{tmpl_s!r}'. It may not be "
            f'a template.{os.linesep}, exc={exc!s}, '
            f'filepath={filepath}, options={options!r}'
        )
        return None

# vim:sw=4:ts=4:et:



#
# Forked from m9dicts.{api,dicts}.
#
# Copyright (C) 2011 - 2021 Red Hat, Inc.
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Utility functions to operate on mapping objects such as get, set and merge.

.. versionadded: 0.8.3
   define _update_* and merge functions based on classes in
   :mod:`m9dicts.dicts`

"""
import collections
import functools
import operator
import re
import typing

from . import utils


# Merge strategies:
MS_REPLACE: str = 'replace'
MS_NO_REPLACE: str = 'noreplace'
MS_DICTS: str = 'merge_dicts'
MS_DICTS_AND_LISTS: str = 'merge_dicts_and_lists'
MERGE_STRATEGIES: typing.Tuple[str, ...] = (
    MS_REPLACE, MS_NO_REPLACE, MS_DICTS, MS_DICTS_AND_LISTS
)

PATH_SEPS: typing.Tuple[str, ...] = ('/', '.')

_JSNP_GET_ARRAY_IDX_REG: typing.Pattern = re.compile(r'(?:0|[1-9][0-9]*)')


DictT = typing.Dict[str, typing.Any]


def _jsnp_unescape(jsn_s: str) -> str:
    """
    Parse and decode given encoded JSON Pointer expression, convert ~1 to
    / and ~0 to ~.

    .. note:: JSON Pointer: http://tools.ietf.org/html/rfc6901

    >>> _jsnp_unescape('/a~1b')
    '/a/b'
    >>> _jsnp_unescape('~1aaa~1~0bbb')
    '/aaa/~bbb'
    """
    return jsn_s.replace('~1', '/').replace('~0', '~')


def _split_path(path: str, seps: typing.Tuple[str, ...] = PATH_SEPS
                ) -> typing.List[str]:
    """
    Parse path expression and return list of path items.

    :param path: Path expression may contain separator chars.
    :param seps: Separator char candidates.
    :return: A list of keys to fetch object[s] later.

    >>> assert _split_path('') == []
    >>> assert _split_path('/') == ['']  # JSON Pointer spec expects this.
    >>> for p in ('/a', '.a', 'a', 'a.'):
    ...     assert _split_path(p) == ['a'], p
    >>> assert _split_path('/a/b/c') == _split_path('a.b.c') == ['a', 'b', 'c']
    >>> assert _split_path('abc') == ['abc']
    """
    if not path:
        return []

    for sep in seps:
        if sep in path:
            if path == sep:  # Special case, '/' or '.' only.
                return ['']
            return [x for x in path.split(sep) if x]

    return [path]


def mk_nested_dic(path: str, val: typing.Any,
                  seps: typing.Tuple[str, ...] = PATH_SEPS) -> DictT:
    """
    Make a nested dict iteratively.

    :param path: Path expression to make a nested dict
    :param val: Value to set
    :param seps: Separator char candidates

    >>> mk_nested_dic('a.b.c', 1)
    {'a': {'b': {'c': 1}}}
    >>> mk_nested_dic('/a/b/c', 1)
    {'a': {'b': {'c': 1}}}
    """
    ret: DictT = dict()
    for key in reversed(_split_path(path, seps)):
        ret = {key: val if not ret else ret.copy()}

    return ret


def get(dic: DictT, path: str, seps: typing.Tuple[str, ...] = PATH_SEPS,
        idx_reg: typing.Pattern = _JSNP_GET_ARRAY_IDX_REG
        ) -> typing.Tuple[typing.Any, str]:
    """getter for nested dicts.

    :param dic: a dict[-like] object
    :param path: Path expression to point object wanted
    :param seps: Separator char candidates
    :return: A tuple of (result_object, error_message)

    >>> d = {'a': {'b': {'c': 0, 'd': [1, 2]}}, '': 3}
    >>> assert get(d, '/') == (3, '')  # key becomes '' (empty string).
    >>> assert get(d, '/a/b/c') == (0, '')
    >>> sorted(get(d, 'a.b')[0].items())
    [('c', 0), ('d', [1, 2])]
    >>> (get(d, 'a.b.d'), get(d, '/a/b/d/1'))
    (([1, 2], ''), (2, ''))
    >>> get(d, "a.b.key_not_exist")  # doctest: +ELLIPSIS
    (None, "'...'")
    >>> get(d, '/a/b/d/2')
    (None, 'list index out of range')
    >>> get(d, '/a/b/d/-')  # doctest: +ELLIPSIS
    (None, 'list indices must be integers...')
    """
    items = [_jsnp_unescape(s) for s in _split_path(path, seps)]  # : [str]
    if not items:
        return (dic, '')
    try:
        if len(items) == 1:
            return (dic[items[0]], '')

        prnt: typing.Any = functools.reduce(operator.getitem, items[:-1], dic)
        arr = (idx_reg.match(items[-1])
               if utils.is_list_like(prnt) else False)  # type: ignore

        return (prnt[int(items[-1])], '') if arr else (prnt[items[-1]], '')

    except (TypeError, KeyError, IndexError) as exc:
        return (None, str(exc))


def set_(dic: DictT, path: str, val: typing.Any,
         seps: typing.Tuple[str, ...] = PATH_SEPS) -> None:
    """setter for nested dicts.

    :param dic: a dict[-like] object support recursive merge operations
    :param path: Path expression to point object wanted
    :param seps: Separator char candidates

    >>> d = dict(a=1, b=dict(c=2, ))
    >>> set_(d, 'a.b.d', 3)
    >>> d['a']['b']['d']
    3
    """
    merge(dic, mk_nested_dic(path, val, seps), ac_merge=MS_DICTS)


def _are_list_like(*objs: typing.Any) -> bool:
    """
    >>> _are_list_like([], (), [x for x in range(10)], (x for x in range(4)))
    True
    >>> _are_list_like([], {})
    False
    >>> _are_list_like([], 'aaa')
    False
    """
    return all(utils.is_list_like(obj) for obj in objs)


def _update_with_replace(self: DictT, other: DictT, key: str,
                         default: typing.Any = None, **_options) -> None:
    """
    Replace value of a mapping object 'self' with 'other' has if both have same
    keys on update. Otherwise, just keep the value of 'self'.

    :param self: mapping object to update with 'other'
    :param other: mapping object to update 'self'
    :param key: key of mapping object to update
    :param val: value to update self alternatively

    :return: None but 'self' will be updated
    """
    oval = other.get(key, None)
    if oval is not None:
        self[key] = oval
    elif default is not None:
        self[key] = default


def _update_wo_replace(self: DictT, other: DictT, key: str,
                       val: typing.Any = None, **_options) -> None:
    """
    Never update (replace) the value of 'self' with 'other''s, that is, only
    the values 'self' does not have its key will be added on update.

    :param self: mapping object to update with 'other'
    :param other: mapping object to update 'self'
    :param key: key of mapping object to update
    :param val: value to update self alternatively

    :return: None but 'self' will be updated
    """
    if key not in self:
        self[key] = other.get(key, val)


def _merge_list(self: DictT, key: str,
                lst: typing.Iterable[typing.Any]) -> None:
    """
    :param key: self[key] will be updated
    :param lst: Other list to merge
    """
    self[key] += [x for x in lst if x not in self[key]]


def _merge_other(self: DictT, key: str, val: typing.Any) -> None:
    """
    :param key: self[key] will be updated
    :param val: Other val to merge (update/replace)
    """
    self[key] = val  # Just overwrite it by default implementation.


def _update_with_merge(self: DictT, other: DictT, key: str,
                       val: typing.Any = None,
                       merge_lists: bool = False, **options) -> None:
    """
    Merge the value of self with other's recursively. Behavior of merge will be
    vary depends on types of original and new values.

    - mapping vs. mapping -> merge recursively
    - list vs. list -> vary depends on 'merge_lists'. see its description.

    :param other: a dict[-like] object or a list of (key, value) tuples
    :param key: key of mapping object to update
    :param val: value to update self[key]
    :param merge_lists:
        Merge not only dicts but also lists. For example,

        [1, 2, 3], [3, 4] ==> [1, 2, 3, 4]
        [1, 2, 2], [2, 4] ==> [1, 2, 2, 4]

    :return: None but 'self' will be updated
    """
    if val is None:
        val = other[key]

    if key in self:
        val0 = self[key]  # Original value
        if utils.is_dict_like(val0):  # It needs recursive updates.
            merge(self[key], val, merge_lists=merge_lists, **options)
        elif merge_lists and _are_list_like(val, val0):
            _merge_list(self, key, val)
        else:
            _merge_other(self, key, val)
    else:
        self[key] = val


def _update_with_merge_lists(self: DictT, other: DictT, key: str,
                             val: typing.Any = None, **options) -> None:
    """
    Similar to _update_with_merge but merge lists always.

    :param self: mapping object to update with 'other'
    :param other: mapping object to update 'self'
    :param key: key of mapping object to update
    :param val: value to update self alternatively

    :return: None but 'self' will be updated
    """
    _update_with_merge(self, other, key, val=val, merge_lists=True, **options)


_MERGE_FNS = {MS_REPLACE: _update_with_replace,
              MS_NO_REPLACE: _update_wo_replace,
              MS_DICTS: _update_with_merge,
              MS_DICTS_AND_LISTS: _update_with_merge_lists}


def _get_update_fn(strategy: str) -> typing.Callable[..., None]:
    """
    Select dict-like class based on merge strategy and orderness of keys.

    :param merge: Specify strategy from MERGE_STRATEGIES of how to merge dicts.
    :return: Callable to update objects
    """
    if strategy is None:
        strategy = MS_DICTS
    try:
        return typing.cast(typing.Callable[..., None], _MERGE_FNS[strategy])
    except KeyError as exc:
        if callable(strategy):
            return strategy

        raise ValueError(f'Wrong merge strategy: {strategy!r}') from exc


UpdatesT = typing.Union[
    typing.Iterable[typing.Tuple[str, typing.Any]],
    DictT
]


def merge(self: DictT, other: UpdatesT, ac_merge: str = MS_DICTS,
          **options) -> None:
    """
    Update (merge) a mapping object 'self' with other mapping object or an
    iterable yields (key, value) tuples based on merge strategy 'ac_merge'.

    :param others: a list of dict[-like] objects or (key, value) tuples
    :param another: optional keyword arguments to update self more
    :param ac_merge: Merge strategy to choose
    """
    _update_fn = _get_update_fn(ac_merge)

    if isinstance(other, dict):
        for key in other.keys():
            _update_fn(self, other, key, **options)
    else:
        try:
            iother = typing.cast(
                typing.Iterable[typing.Tuple[str, typing.Any]],
                other
            )
            for key, val in iother:
                _update_fn(self, dict(other), key, val=val, **options)
        except (ValueError, TypeError) as exc:  # Re-raise w/ info.
            raise type(exc)(f'{exc!s} other={other!r}')


def _make_recur(obj: typing.Any, make_fn: typing.Callable,
                ac_ordered: bool = False,
                ac_dict: typing.Optional[typing.Callable] = None,
                **options) -> DictT:
    """
    :param obj: A mapping objects or other primitive object
    :param make_fn: Function to make/convert to
    :param ac_ordered: Use OrderedDict instead of dict to keep order of items
    :param ac_dict: Callable to convert 'obj' to mapping object
    :param options: Optional keyword arguments.

    :return: Mapping object
    """
    if ac_dict is None:
        ac_dict = collections.OrderedDict if ac_ordered else dict

    return ac_dict((k, None if v is None else make_fn(v, **options))
                   for k, v in obj.items())


def _make_iter(obj: typing.Any, make_fn: typing.Callable, **options
               ) -> DictT:
    """
    :param obj: A mapping objects or other primitive object
    :param make_fn: Function to make/convert to
    :param options: Optional keyword arguments.

    :return: Mapping object
    """
    return type(obj)(make_fn(v, **options) for v in obj)


def convert_to(obj: typing.Any, ac_ordered: bool = False,
               ac_dict: typing.Optional[typing.Callable] = None,
               **options) -> DictT:
    """
    Convert a mapping objects to a dict or object of 'to_type' recursively.
    Borrowed basic idea and implementation from bunch.unbunchify. (bunch is
    distributed under MIT license same as this.)

    :param obj: A mapping objects or other primitive object
    :param ac_ordered: Use OrderedDict instead of dict to keep order of items
    :param ac_dict: Callable to convert 'obj' to mapping object
    :param options: Optional keyword arguments.

    :return: A dict or OrderedDict or object of 'cls'

    >>> OD = collections.OrderedDict
    >>> convert_to(OD((('a', 1) ,)), cls=dict)
    {'a': 1}
    >>> convert_to(OD((('a', OD((('b', OD((('c', 1), ))), ))), )), cls=dict)
    {'a': {'b': {'c': 1}}}
    """
    options.update(ac_ordered=ac_ordered, ac_dict=ac_dict)
    if utils.is_dict_like(obj):
        return _make_recur(obj, convert_to, **options)
    if utils.is_list_like(obj):
        return _make_iter(obj, convert_to, **options)

    return obj

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Abstract and basic loaders.
"""
import collections
import io
import pathlib
import typing

from ...utils import (
    filter_options, is_stream_ioinfo
)
from .datatypes import (
    InDataExT, IoiT, GenContainerT, OptionsT
)
from .utils import not_implemented


DATA_DEFAULT: InDataExT = dict()


class LoaderMixin:
    """
    Mixin class to load data.

    Inherited classes must implement the following methods.

    - :meth:`load_from_string`: Load config from string
    - :meth:`load_from_stream`: Load config from a file or file-like object
    - :meth:`load_from_path`: Load config from file of given path

    Member variables:

    - _load_opts: Backend specific options on load
    - _ordered: True if the parser keep the order of items by default
    - _allow_primitives: True if the parser.load* may return objects of
      primitive data types other than mapping types such like JSON parser
    - _dict_opts: Backend options to customize dict class to make results
    - _open_read_mode: Backend option to specify read mode passed to open()
    """
    _load_opts: typing.List[str] = []
    _ordered: bool = False
    _allow_primitives: bool = False
    _dict_opts: typing.List[str] = []
    _open_read_mode: str = 'r'

    @classmethod
    def ordered(cls) -> bool:
        """
        :return: True if parser can keep the order of keys else False.
        """
        return cls._ordered

    @classmethod
    def allow_primitives(cls) -> bool:
        """
        :return:
            True if the parser.load* may return objects of primitive data types
            other than mapping types such like JSON parser
        """
        return cls._allow_primitives

    @classmethod
    def dict_options(cls) -> typing.List[str]:
        """
        :return: List of dict factory options
        """
        return cls._dict_opts

    def ropen(self, filepath, **kwargs):
        """
        :param filepath: Path to file to open to read data
        """
        return open(  # pylint: disable=consider-using-with
            filepath, self._open_read_mode, **kwargs
        )

    def _container_factory(self, **options) -> GenContainerT:
        """
        The order of prirorities are ac_dict, backend specific dict class
        option, ac_ordered.

        :param options: Keyword options may contain 'ac_ordered'.
        :return: Factory (class or function) to make an container.
        """
        ac_dict = options.get("ac_dict", False)
        _dicts = [x for x in (options.get(o) for o in self.dict_options())
                  if x]

        if self.dict_options() and ac_dict and callable(ac_dict):
            return ac_dict  # Higher priority than ac_ordered.
        if _dicts and callable(_dicts[0]):
            return _dicts[0]
        if self.ordered() and options.get("ac_ordered", False):
            return collections.OrderedDict

        return dict

    def _load_options(self, container: GenContainerT, **options) -> OptionsT:
        """
        Select backend specific loading options.
        """
        # Force set dict option if available in backend. For example,
        # options["object_hook"] will be OrderedDict if 'container' was
        # OrderedDict in JSON backend.
        for opt in self.dict_options():
            options.setdefault(opt, container)

        return filter_options(self._load_opts, options)

    def load_from_string(self, content: str, container: GenContainerT,
                         **kwargs) -> InDataExT:
        """
        Load config from given string 'content'.

        :param content: Config content string
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        not_implemented(self, content, container, **kwargs)
        return DATA_DEFAULT

    def load_from_path(self, filepath: str, container: GenContainerT,
                       **kwargs) -> InDataExT:
        """
        Load config from given file path 'filepath`.

        :param filepath: Config file path
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        not_implemented(self, filepath, container, **kwargs)
        return DATA_DEFAULT

    def load_from_stream(self, stream: typing.IO, container: GenContainerT,
                         **kwargs) -> InDataExT:
        """
        Load config from given file like object 'stream`.

        :param stream:  Config file or file like object
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        not_implemented(self, stream, container, **kwargs)
        return DATA_DEFAULT

    def loads(self, content: str, **options) -> InDataExT:
        """
        Load config from given string 'content' after some checks.

        :param content:  Config file content
        :param options:
            options will be passed to backend specific loading functions.
            please note that options have to be sanitized w/
            :func:`anyconfig.utils.filter_options` later to filter out options
            not in _load_opts.

        :return: dict or dict-like object holding configurations
        """
        container = self._container_factory(**options)
        if not content or content is None:
            return container()

        options = self._load_options(container, **options)
        return self.load_from_string(content, container, **options)

    def load(self, ioi: IoiT, ac_ignore_missing: bool = False,
             **options) -> InDataExT:
        """
        Load config from a file path or a file / file-like object which 'ioi'
        referring after some checks.

        :param ioi:
            'anyconfig.common.IOInfo' namedtuple object provides various info
            of input object to load data from

        :param ac_ignore_missing:
            Ignore and just return empty result if given `ioi` object does not
            exist in actual.
        :param options:
            options will be passed to backend specific loading functions.
            please note that options have to be sanitized w/
            :func:`anyconfig.utils.filter_options` later to filter out options
            not in _load_opts.

        :return: dict or dict-like object holding configurations
        """
        container = self._container_factory(**options)
        options = self._load_options(container, **options)

        if not ioi:
            return container()

        if is_stream_ioinfo(ioi):
            cnf = self.load_from_stream(
                typing.cast(typing.IO, ioi.src), container, **options
            )
        else:
            if ac_ignore_missing and not pathlib.Path(ioi.path).exists():
                return container()

            cnf = self.load_from_path(ioi.path, container, **options)

        return cnf


class BinaryLoaderMixin(LoaderMixin):
    """
    Mixin class to load binary (byte string) configuration files.
    """
    _open_read_mode: str = 'rb'


class FromStringLoaderMixin(LoaderMixin):
    """
    Abstract config parser provides a method to load configuration from string
    content to help implement parser of which backend lacks of such function.

    Parser classes inherit this class have to override the method
    :meth:`load_from_string` at least.
    """
    def load_from_stream(self, stream: typing.IO, container: GenContainerT,
                         **kwargs) -> InDataExT:
        """
        Load config from given stream 'stream'.

        :param stream: Config file or file-like object
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        return self.load_from_string(stream.read(), container, **kwargs)

    def load_from_path(self, filepath: str, container: GenContainerT,
                       **kwargs) -> InDataExT:
        """
        Load config from given file path 'filepath'.

        :param filepath: Config file path
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        with self.ropen(filepath) as inp:
            return self.load_from_stream(inp, container, **kwargs)


class FromStreamLoaderMixin(LoaderMixin):
    """
    Abstract config parser provides a method to load configuration from string
    content to help implement parser of which backend lacks of such function.

    Parser classes inherit this class have to override the method
    :meth:`load_from_stream` at least.
    """
    def load_from_string(self, content: str, container: GenContainerT,
                         **kwargs) -> InDataExT:
        """
        Load config from given string 'cnf_content'.

        :param content: Config content string
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        return self.load_from_stream(io.StringIO(content),
                                     container, **kwargs)

    def load_from_path(self, filepath: str, container: GenContainerT,
                       **kwargs) -> InDataExT:
        """
        Load config from given file path 'filepath'.

        :param filepath: Config file path
        :param container: callble to make a container object later
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        with self.ropen(filepath) as inp:
            return self.load_from_stream(inp, container, **kwargs)

# vim:sw=4:ts=4:et:



# -*- coding: utf-8 -*-
#
# pylint:disable=invalid-name
import sys
import pathlib

sys.path.insert(0, str(pathlib.Path(__file__).parent.resolve() / 'src'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_autodoc_typehints'
]
source_suffix = '.rst'
master_doc = 'index'

project = u'python-anyconfig'
copyright = u'2021, Satoru SATOH <satoru.satoh@gmail.com>'
version = '0.10.0'
release = version

exclude_patterns = []

html_theme = 'default'

autodoc_member_order = 'bysource'



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Misc utility functions.
"""
import typing


def noop(val: typing.Any, *_args, **_kwargs) -> typing.Any:
    """A function does nothing.

    >>> noop(1)
    1
    """
    return val


def filter_options(keys: typing.Iterable[str],
                   options: typing.Mapping[str, typing.Any]
                   ) -> typing.Dict[str, typing.Any]:
    """
    Filter 'options' with given 'keys'.

    :param keys: key names of optional keyword arguments
    :param options: optional keyword arguments to filter with 'keys'

    >>> filter_options(("aaa", ), dict(aaa=1, bbb=2))
    {'aaa': 1}
    >>> filter_options(("aaa", ), dict(bbb=2))
    {}
    """
    return dict((k, options[k]) for k in keys if k in options)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring, invalid-name
import unittest

import anyconfig.ioinfo
import anyconfig.processors.utils as TT

from .common import A, A2, A3, B, C, PRS


PRS = [p() for p in PRS]  # Instantiate all.


class TestCase(unittest.TestCase):

    def test_select_by_key(self):
        ies = (([], []),
               (((['a'], 1), ), [('a', [1])]),
               (((['a', 'aaa'], 1),
                 (['b', 'bb'], 2),
                 (['a'], 3)),
                [('a', [1, 3]),
                 ('aaa', [1]),
                 ('b', [2]),
                 ('bb', [2])]))

        for inp, exp in ies:
            self.assertEqual(TT.select_by_key(inp), exp)

    def test_select_by_key_reversed(self):
        ies = ((((['a', 'aaa'], 1),
                 (['a'], 3)),
                [('a', [3, 1]),
                 ('aaa', [1])]),
               )

        def sfn(itr):
            return sorted(itr, reverse=True)

        for inp, exp in ies:
            self.assertEqual(TT.select_by_key(inp, sfn), exp)

    def test_list_by_x(self):
        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        ies = ((([], 'type'), []),
               (([a], 'type'), [(a.type(), [a])]),
               (([a], 'extensions'),
                [(x, [a]) for x in a.extensions()]),
               (((a, a2, a3), 'type'),
                [(a.type(), [a3, a2, a])]),
               (([a, b, c], 'type'),
                [(a.type(), [a]), (b.type(), [b, c])]),
               ((PRS, 'type'),
                [(a.type(), [a3, a2, a]), (b.type(), [b, c])]),
               ((PRS, 'extensions'),
                [('js', [a3, a2, a]), ('json', [a3, a2, a]),
                 ('jsn', [a3, a2, a]), ('yaml', [b, c]), ('yml', [b, c])]),
               )

        for prs_key, exp in ies:
            self.assertEqual(
                sorted(TT.list_by_x(*prs_key)), sorted(exp)
            )

    def test_list_by_x_ng_cases(self):
        with self.assertRaises(ValueError):
            TT.list_by_x(PRS, 'undef')

    def test_findall_with_pred__type(self):
        def _findall_by_type(typ):
            return TT.findall_with_pred(lambda p: p.type() == typ, PRS)

        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        ies = (('json', [a3, a2, a]),
               ('yaml', [b, c]),
               ('undefined', []),
               )

        for inp, exp in ies:
            self.assertEqual(_findall_by_type(inp), exp)

    def test_findall_with_pred__ext(self):
        def _findall_with_pred__ext(ext):
            return TT.findall_with_pred(lambda p: ext in p.extensions(), PRS)

        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        ies = (('js', [a3, a2, a]),
               ('yml', [b, c]),
               ('xyz', []),
               )

        for inp, exp in ies:
            self.assertEqual(_findall_with_pred__ext(inp), exp)

    def assertInstance(self, obj, cls):
        self.assertTrue(isinstance(obj, cls))

    def test_maybe_processor(self):
        a3 = A3()
        ies = (((a3, A3), True),
               ((A3, A3), True),
               ((B, A3), False),
               )
        for inp, exp in ies:
            afn = self.assertTrue if exp else self.assertFalse
            res = TT.maybe_processor(*inp)
            afn(isinstance(res, A3))

            if not exp:
                self.assertTrue(res is None)

    def test_find_by_type_or_id(self):
        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        ies = ((('json', PRS), [a3, a2, a]),
               (('yaml', PRS), [b, c]),
               (('dummy', PRS), [c]),
               )
        for inp, exp in ies:
            self.assertEqual(TT.find_by_type_or_id(*inp), exp)

    def test_find_by_type_or_id_ng_cases(self):
        with self.assertRaises(TT.UnknownProcessorTypeError):
            TT.find_by_type_or_id('xyz', PRS)

    def test_find_by_fileext(self):
        ies = ((('js', PRS), [A3(), A2(), A()]),
               (('yml', PRS), [B(), C()]),
               )
        for inp, exp in ies:
            self.assertEqual(TT.find_by_fileext(*inp), exp)

    def test_find_by_fileext_ng_cases(self):
        with self.assertRaises(TT.UnknownFileTypeError):
            TT.find_by_fileext('xyz', PRS)

    def test_find_by_maybe_file(self):
        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        obj = anyconfig.ioinfo.make('/path/to/a.json')

        ies = ((('/path/to/a.jsn', PRS), [a3, a2, a]),
               (('../../path/to/b.yml', PRS), [b, c]),
               ((obj, PRS), [a3, a2, a]),
               )

        for inp, exp in ies:
            self.assertEqual(TT.find_by_maybe_file(*inp), exp)

    def test_find_by_maybe_file_ng_cases(self):
        ies = (('/tmp/x.xyz', PRS),
               ('/dev/null', PRS),
               )
        for inp in ies:
            with self.assertRaises(TT.UnknownFileTypeError):
                TT.find_by_maybe_file(*inp)

    def test_findall_ng_cases(self):
        ies = (
               ((None, PRS, None), ValueError),  # w/o path nor type
               (('/tmp/x.xyz', PRS, None), TT.UnknownFileTypeError),
               (('/dev/null', PRS, None), TT.UnknownFileTypeError),
               ((None, PRS, 'xyz'), TT.UnknownProcessorTypeError),
               )
        for inp, exc in ies:
            with self.assertRaises(exc):
                TT.findall(*inp)

    def test_findall_by_maybe_file(self):
        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        obj = anyconfig.ioinfo.make('/path/to/a.json')

        ies = ((('/path/to/a.jsn', PRS), [a3, a2, a]),
               (('../../path/to/b.yml', PRS), [b, c]),
               ((obj, PRS), [a3, a2, a]),
               )
        for inp, exp in ies:
            self.assertEqual(TT.findall(*inp), exp)

    def test_findall_by_type_or_id(self):
        (a, a2, a3, b, c) = (A(), A2(), A3(), B(), C())
        ies = (((None, PRS, 'json'), [a3, a2, a]),
               ((None, PRS, 'yaml'), [b, c]),
               ((None, PRS, 'dummy'), [c]),
               )
        for inp, exp in ies:
            self.assertEqual(TT.findall(*inp), exp)

    def test_find_by_forced_type(self):
        a2 = A2()
        c = C()
        ies = (((None, PRS, A2), a2),
               ((None, PRS, A2), a2),
               ((None, PRS, c.cid()), c),
               )

        for inp, exp in ies:
            self.assertEqual(TT.find(*inp), exp)

    def test_find__maybe_file(self):
        (a3, b) = (A3(), B())
        obj = anyconfig.ioinfo.make('/path/to/a.json')

        ies = ((('/path/to/a.jsn', PRS), a3),
               (('../../path/to/b.yml', PRS), b),
               ((obj, PRS), a3),
               )
        for inp, exp in ies:
            self.assertEqual(TT.find(*inp), exp)

    def test_find__type_or_id(self):
        ies = (((None, PRS, 'json'), A3()),
               ((None, PRS, 'yaml'), B()),
               ((None, PRS, 'dummy'), C()),
               )
        for inp, exp in ies:
            self.assertEqual(TT.find(*inp), exp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import pathlib
import tempfile

import anyconfig.api._dump as TT

from anyconfig.api import (
    UnknownProcessorTypeError, UnknownFileTypeError
)

from . import common


class TestCase(common.BaseTestCase):

    def test_dump(self):
        with tempfile.TemporaryDirectory() as tdir:
            for data in self.each_data():
                out = pathlib.Path(tdir) / 'out.json'
                TT.dump(data.inp, out, **data.opts)
                self.assertEqual(
                    out.read_text().strip(),
                    data.exp.strip(),
                    f'{data.datadir!s}, {data.inp_path!s}'
                )

    def test_dump_intentional_failures(self):
        with tempfile.TemporaryDirectory() as tdir:
            for data in self.each_data():
                out = pathlib.Path(tdir) / 'out.json'
                TT.dump(data.inp, out, **data.opts)
                with self.assertRaises(AssertionError):
                    self.assertEqual(out.read_text().strip(), '')

    def test_dump_failure_ac_parser_was_not_given(self):
        for data in self.each_data():
            with self.assertRaises(UnknownFileTypeError):
                TT.dump(data.inp, 'dummy.txt')

    def test_dump_failure_invalid_ac_parser_was_given(self):
        for data in self.each_data():
            with self.assertRaises(UnknownProcessorTypeError):
                TT.dump(data.inp, 'dummy.json', ac_parser='invalid_id')

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=too-few-public-methods
r"""Singleton class

.. versionadded:: 0.9.8

   - Add to make a kind of manager instancne later to manage plugins.
"""
import threading


class Singleton:
    """Singleton utilizes __new__ special method.

    .. note:: Inherited classes are equated with base class inherit this.
    """
    __instance = None
    __lock = threading.RLock()

    def __new__(cls):
        if cls.__instance is None:
            with cls.__lock:
                if cls.__instance is None:
                    cls.__instance = object.__new__(cls)

        return cls.__instance

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

from ... import base


class BaseTestCase(base.TDataCollector, unittest.TestCase):
    target = 'dump'

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Test cases for anyconfig.utils.files.
"""
import pathlib
import unittest

import anyconfig.ioinfo.utils as TT


class TestCase(unittest.TestCase):

    def test_is_io_stream(self):
        ies = (
            (open(__file__), True),
            (__file__, False),
            ([__file__], False),
            (pathlib.Path(__file__), False),
            ([pathlib.Path(__file__)], False),
        )
        for inp, exp in ies:
            res = TT.is_io_stream(inp)
            (self.assertTrue if exp else self.assertFalse)(res)

    def test_get_path_and_ext(self):
        this = pathlib.Path(__file__)
        ies = (
            (this, (this.resolve(), 'py')),
        )
        for inp, exp in ies:
            res = TT.get_path_and_ext(inp)
            self.assertEqual(res, exp)

    def test_split_path_by_marker(self):
        ies = (
            ('a.txt', ('a.txt', '')),
            ('*.txt', ('', '*.txt')),
            ('a/*.txt', ('a', '*.txt')),
            ('a/b/*.txt', ('a/b', '*.txt')),
            ('a/b/*/*.txt', ('a/b', '*/*.txt')),
        )
        for inp, exp in ies:
            self.assertEqual(TT.split_path_by_marker(inp), exp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-import,import-error,invalid-name
r"""Public APIs of anyconfig module.

.. versionchanged:: 0.10.2

   - Re-structured APIs and split into sub modules
   - Added type hints

.. versionchanged:: 0.9.9

   - Removed the API 'find_loader'
   - Added new APIs :func:`find` and :func:`findall` to :func:`find parsers`
     (loaders and dumpers) to suppport to find multiple parsers, and replace
     the API 'find_loader'
   - Added new APIs :func:`list_by_cid`, :func:`list_by_type` and
     :func:`list_by_extension` to list parsers by various viewpoints.

.. versionadded:: 0.9.8

   - Added new API load_plugins to [re-]load plugins

.. versionadded:: 0.9.5

   - Added pathlib support. Now all of load and dump APIs can process
     pathlib.Path object basically.
   - 'ignore_missing' keyword option for load APIs are now marked as deprecated
     and will be removed soon.
   - Allow to load data other than mapping objects for some backends such as
     JSON and YAML.

.. versionadded:: 0.8.3

   - Added ac_dict keyword option to pass dict factory (any callable like
     function or class) to make dict-like object in backend parsers.
   - Added ac_query keyword option to query data with JMESPath expression.
   - Added experimental query api to query data with JMESPath expression.
   - Removed ac_namedtuple keyword option.
   - Export :func:`merge`.
   - Stop exporting :func:`to_container` which was deprecated and removed.

.. versionadded:: 0.8.2

   - Added new API, version to provide version information.

.. versionadded:: 0.8.0

   - Removed set_loglevel API as it does not help much.
   - Added :func:`open` API to open files with appropriate open mode.
   - Added custom exception classes, :class:`UnknownProcessorTypeError` and
     :class:`UnknownFileTypeError` to express specific errors.
   - Change behavior of the API :func:`find_loader` and others to make them
     fail firt and raise exceptions (ValueError, UnknownProcessorTypeError or
     UnknownFileTypeError) as much as possible if wrong parser type for uknown
     file type was given.

.. versionadded:: 0.5.0

   - Most keyword arguments passed to APIs are now position independent.
   - Added ac_namedtuple parameter to \*load and \*dump APIs.

.. versionchanged:: 0.3

   - Replaced 'forced_type' optional argument of some public APIs with
     'ac_parser' to allow skip of config parser search by passing parser object
     previously found and instantiated.

     Also removed some optional arguments, 'ignore_missing', 'merge' and
     'marker', from definitions of some public APIs as these may not be changed
     from default in common use cases.

.. versionchanged:: 0.2

   - Now APIs :func:`find_loader`, :func:`single_load`, :func:`multi_load`,
     :func:`load` and :func:`dump` can process a file/file-like object or a
     list of file/file-like objects instead of a file path or a list of file
     paths.

.. versionadded:: 0.2

   - Export factory method (create) of anyconfig.mergeabledict.MergeableDict
"""
from .datatypes import MaybeDataT
from ._dump import (
    dump, dumps
)
from ._load import (
    single_load, multi_load, load, loads
)
from ._open import open  # pylint: disable=redefined-builtin

# Export some more APIs originally from other sub modules.
from ..backend import ParserT
from ..common import (
    IOInfo, InDataT, InDataExT,
    UnknownFileTypeError, UnknownParserTypeError,
    UnknownProcessorTypeError, ValidationError
)
from ..dicts import (
    MS_REPLACE, MS_NO_REPLACE, MS_DICTS, MS_DICTS_AND_LISTS, MERGE_STRATEGIES,
    merge, get, set_,
)
from ..parsers import (
    load_plugins, list_types, list_by_cid, list_by_type, list_by_extension,
    findall, find, MaybeParserT
)
from ..query import try_query
from ..schema import (
    validate, is_valid, gen_schema
)


__version__ = '0.11.1'


def version():
    """
    :return: A tuple of version info, (major, minor, release), e.g. (0, 8, 2)
    """
    return __version__.split('.')


__all__ = [
    'MaybeDataT',
    'dump', 'dumps',
    'single_load', 'multi_load', 'load', 'loads',
    'open', 'version',

    # anyconfig.backend
    'ParserT',

    # anyconfig.common
    'IOInfo', 'InDataT', 'InDataExT',
    'UnknownFileTypeError', 'UnknownParserTypeError',
    'UnknownProcessorTypeError', 'ValidationError',

    # anyconfig.dicsts
    'MS_REPLACE', 'MS_NO_REPLACE', 'MS_DICTS', 'MS_DICTS_AND_LISTS',
    'MERGE_STRATEGIES',
    'merge', 'get', 'set_',

    # anyconfig.parsers
    'load_plugins', 'list_types', 'list_by_cid', 'list_by_type',
    'list_by_extension', 'findall', 'find',
    'MaybeParserT',

    # anyconfig.query
    'try_query',

    # anyconfig.validate
    'validate', 'is_valid', 'gen_schema'
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import anyconfig.api._load as TT


class Basa:
    @staticmethod
    def target_fn(*args, **kwargs):
        return TT.load(*args, **kwargs)


class MultiBase:
    target: str = 'load/multi'


class SingleBase:
    target: str = 'load/single'

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring, invalid-name
import unittest

import anyconfig.ioinfo.factory as TT

from anyconfig.common import (
    IOInfo, IOI_PATH_OBJ, IOI_STREAM
)

from .constants import TEST_PY


TEST_IOI_PATH_OBJ = IOInfo(
    src=TEST_PY, type=IOI_PATH_OBJ, path=str(TEST_PY), extension='py'
)
TEST_IOI_STREAM = IOInfo(
    src=TEST_PY.open(), type=IOI_STREAM, path=str(TEST_PY), extension='py'
)


class TestCase(unittest.TestCase):

    def test_make(self):
        ies = (
            (TEST_IOI_PATH_OBJ, TEST_IOI_PATH_OBJ),
            (TEST_IOI_STREAM, TEST_IOI_STREAM),
            (str(TEST_PY), TEST_IOI_PATH_OBJ),
        )
        for inp, exp in ies:
            self.assertEqual(TT.make(inp), exp)

    def test_make_failures(self):
        inps = (None, )
        for inp in inps:
            with self.assertRaises(ValueError):
                TT.make(inp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

import anyconfig.api._load as TT

from ... import base


class Collector(base.TDataCollector):

    @staticmethod
    def target_fn(*args, **kwargs):
        return TT.single_load(*args, **kwargs)


class TestCase(unittest.TestCase, Collector):

    def setUp(self):
        self.init()

    def test_single_load(self):
        for data in self.each_data():
            self.assertEqual(
                self.target_fn(data.inp_path, **data.opts),
                data.exp,
                data
            )

    def test_single_load_intentional_failures(self):
        for data in self.each_data():
            with self.assertRaises(AssertionError):
                self.assertEqual(
                    self.target_fn(data.inp_path, **data.opts),
                    None
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
from . import common


class TestCase(common.TestCase):
    kind = 'multi_types'
    pattern = '*.*'

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=too-few-public-methods
"""anyconfig basic data types.
"""
import typing


class BaseError(RuntimeError):
    """Base Error exception."""
    _msg_fmt: str = 'forced_type: {!s}'

    def __init__(self, arg: typing.Optional[typing.Any] = None):
        super().__init__(self._msg_fmt.format(str(arg)))


class UnknownParserTypeError(BaseError):
    """Raise if no parsers were found for given type."""
    _msg_fmt: str = 'No parser found for type: {!s}'


class UnknownProcessorTypeError(UnknownParserTypeError):
    """Raise if no processors were found for given type."""


class UnknownFileTypeError(BaseError):
    """Raise if not parsers were found for given file path."""
    _msg_fmt: str = 'No parser found for file: {!s}'


class ValidationError(BaseError):
    """Raise if validation failed."""
    _msg_fmt: str = 'Validation failed: {!s}'

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Globals, functions common in some JSON backends:

Changelog:

.. versionadded:: 0.9.8
"""
from .. import base


JSON_LOAD_OPTS = ['cls', 'object_hook', 'parse_float', 'parse_int',
                  'parse_constant', 'object_pairs_hook']

JSON_DUMP_OPTS = ['skipkeys', 'ensure_ascii', 'check_circular', 'allow_nan',
                  'cls', 'indent', 'separators', 'default', 'sort_keys']

JSON_DICT_OPTS = ['object_pairs_hook', 'object_hook']


class Parser(base.StringStreamFnParser):
    """
    Parser for JSON files.
    """
    _cid = 'std.json'
    _type = 'json'
    _extensions = ['json', 'jsn', 'js']
    _ordered = True
    _allow_primitives = True

    # .. note:: These may be overwritten.
    _load_opts = JSON_LOAD_OPTS
    _dump_opts = JSON_DUMP_OPTS
    _dict_opts = JSON_DICT_OPTS

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
"""File based test data collector.
"""
import inspect
import pathlib
import typing

from . import (
    constants, datatypes, utils
)


DICT_0 = dict()


class TDataCollector:
    """File based test data collector.
    """
    target: str = ''  # Initial value will be replaced in self.init.
    kind: str = 'basics'
    pattern: str = '*.json'  # input file name pattern

    # sub dir names of expected data files should be found always.
    should_exist: typing.Iterable[str] = ('e', )

    # True if you want to keep the order of keys of dicts loaded.
    ordered: bool = False

    root: typing.Optional[pathlib.Path] = None
    datasets: typing.List[datatypes.TData] = []
    initialized: bool = False

    @classmethod
    def resolve_target(cls) -> str:
        """
        Resolve target by this file path.
        """
        return utils.target_by_parent(inspect.getfile(cls))

    def init(self) -> None:
        """Initialize its members.
        """
        if not self.target:
            self.target = self.resolve_target()

        if not self.root:
            self.root = constants.RES_DIR / self.target / self.kind

        self.datasets = self.load_datasets()
        self.initialized = True

    def load_datasets(self) -> typing.List[datatypes.TData]:
        """Load test data from files.
        """
        _datasets = [
            (datadir,
             [datatypes.TData(
                data.datadir, data.inp,
                utils.load_data(data.inp, ordered=self.ordered),
                utils.load_data(data.exp, ordered=self.ordered),
                utils.load_data(data.opts, default=DICT_0),
                data.scm,
                utils.load_data(data.query, default=''),
                utils.load_data(data.ctx, default=DICT_0, ordered=self.ordered)
              )
              for data in utils.each_data_from_dir(
                  datadir, self.pattern, self.should_exist
              )]
             )
            for datadir in sorted(self.root.glob('*'))
        ]
        if not _datasets:
            raise ValueError(f'No data: {self.root!s}')

        for datadir, data in _datasets:
            if not data:
                raise ValueError(
                    f'No data in subdir: {datadir!s}, '
                    f'pattern={self.pattern}, '
                    f'should_exist={self.should_exist!r}'
                )

        return _datasets

    def each_data(self) -> typing.Iterable[datatypes.TData]:
        """Yields test data.
        """
        if not self.initialized:
            self.init()

        for _datadir, data in self.datasets:
            for tdata in data:
                yield tdata

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
"""Test cases for anyconfig.backend.base.utils.
"""
import pathlib
import tempfile
import unittest

import anyconfig.backend.base.utils as TT


FILENAME = 'file_not_exist.txt'


class TestCase(unittest.TestCase):

    def test_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            TT.not_implemented()

    def test_ensure_outdir_exists_for_file_if_it_exists(self):
        outdir = pathlib.Path.cwd()
        outfile = outdir / FILENAME

        TT.ensure_outdir_exists(outfile)
        TT.ensure_outdir_exists(str(outfile))
        self.assertTrue(outdir.exists())

    def test_ensure_outdir_exists_for_file_if_it_does_not_exist(self):
        with tempfile.TemporaryDirectory() as outdir:
            outdir = pathlib.Path(outdir)
            outpath = outdir / FILENAME

            TT.ensure_outdir_exists(outpath)
            self.assertTrue(outdir.exists())

    def test_ensure_outdir_exists_for_file_if_its_parent_does_not_exist(self):
        with tempfile.TemporaryDirectory() as parent:
            outdir = pathlib.Path(parent) / 'a' / 'b' / 'c'
            outpath = outdir / FILENAME

            TT.ensure_outdir_exists(outpath)
            self.assertTrue(outdir.exists())

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import warnings

import anyconfig.api._load as TT

from anyconfig.api import UnknownProcessorTypeError

from . import common


class TestCase(common.TestCase):

    def test_loads(self):
        for data in self.each_data():
            self.assertEqual(
                TT.loads(data.inp, **data.opts),
                data.exp,
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_loads_intentional_failures(self):
        for data in self.each_data():
            with self.assertRaises(AssertionError):
                self.assertEqual(TT.loads(data.inp, **data.opts), {})

    def test_loads_failure_ac_parser_was_not_given(self):
        for data in self.each_data():
            with warnings.catch_warnings(record=True) as warns:
                warnings.simplefilter('always')
                self.assertEqual(TT.loads(data.inp), None)
                self.assertEqual(len(warns), 1)
                self.assertTrue(issubclass(warns[-1].category, UserWarning))
                self.assertTrue(
                    'ac_parser was not given but' in str(warns[-1].message)
                )

    def test_loads_failure_invalid_ac_parser_was_given(self):
        for data in self.each_data():
            with self.assertRaises(UnknownProcessorTypeError):
                self.assertEqual(
                    TT.loads(data.inp, ac_parser='invalid_id'),
                    None
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

import anyconfig.api

from . import common


LOADER_TYPES = frozenset(anyconfig.api.list_types())


@unittest.skipIf('yaml' not in LOADER_TYPES,
                 'yaml loader is not available')
class YamlTestCase(common.TestCase):
    kind = 'yaml'
    pattern = '*.yml'


@unittest.skipIf('toml' not in LOADER_TYPES,
                 'toml loader is not available')
class TomlTestCase(YamlTestCase):
    kind = 'toml'
    pattern = '*.toml'

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=import-error
r"""JSON backend by simplejson:

- Format to support: JSON, http://www.json.org
- Requirements: simplejson
- Development Status :: 5 - Production/Stable
- Limitations: None obvious
- Special options:

  - All options of simplejson.load{s,} and simplejson.dump{s,} except
    object_hook should work.

  - See also: https://simplejson.readthedocs.io

Changelog:

.. versionchanged:: 0.9.8

   - Exported from ..json.py
"""
import simplejson as json

from .. import base
from .common import (
    JSON_LOAD_OPTS, JSON_DUMP_OPTS, Parser as BaseParser
)


JSON_LOAD_OPTS.append('use_decimal')
JSON_DUMP_OPTS.extend(['use_decimal', 'namedtuple_as_object', 'tuple_as_array',
                       'bigint_as_string', 'item_sort_key', 'for_json',
                       'ignore_nan', 'int_as_string_bitcount',
                       'iterable_as_array'])


class Parser(BaseParser):
    """
    Parser for JSON files.
    """
    _cid = 'simplejson'
    _load_opts = JSON_LOAD_OPTS
    _dump_opts = JSON_DUMP_OPTS

    _load_from_string_fn = base.to_method(json.loads)
    _load_from_stream_fn = base.to_method(json.load)
    _dump_to_string_fn = base.to_method(json.dumps)
    _dump_to_stream_fn = base.to_method(json.dump)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2015 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, unused-variable, invalid-name
import os
import pathlib
import unittest
import unittest.mock

try:
    import anyconfig.template.jinja2 as TT
except ImportError:
    raise unittest.SkipTest('jinja2 does not look available.')

import tests.common


C_1 = """A char is 'a'.
A char is 'b'.
A char is 'c'.
"""

TMPLS = [('00.j2', "{% include '10.j2' %}" + os.linesep, C_1),
         ('10.j2', """{% for c in ['a', 'b', 'c'] -%}
A char is '{{ c }}'.
{% endfor %}
""", C_1)]

C_2 = "-1"
TMPL_WITH_FILTER = ('11.j2', "{{ 1|negate }}", C_2)


def negate(value):
    return -value


class TestCase(tests.common.TestCaseWithWorkdir):

    templates = TMPLS
    curdir = str(pathlib.Path('..').resolve())

    def setUp(self):
        super().setUp()
        for fname, tmpl_s, _ctx in self.templates + [TMPL_WITH_FILTER]:
            (pathlib.Path(self.workdir) / fname).write_text(tmpl_s)

    def test_10_render_impl__wo_paths(self):
        for fname, _str, ctx in self.templates:
            fpath = pathlib.Path(self.workdir) / fname
            c_r = TT.render_impl(fpath)
            self.assertEqual(c_r, ctx)

    def test_12_render_impl__w_paths(self):
        for fname, _str, ctx in self.templates:
            fpath = pathlib.Path(self.workdir) / fname
            c_r = TT.render_impl(fpath, paths=[self.workdir])
            self.assertEqual(c_r, ctx)

    def test_20_render__wo_paths(self):
        for fname, _str, ctx in self.templates:
            fpath = pathlib.Path(self.workdir) / fname
            c_r = TT.render(str(fpath))
            self.assertEqual(c_r, ctx)

    def test_22_render__w_wrong_tpath(self):
        ng_t = pathlib.Path(self.workdir) / "ng.j2"
        ok_t = pathlib.Path(self.workdir) / "ok.j2"
        ok_t_content = "a: {{ a }}"
        ok_content = "a: aaa"
        ctx = dict(a="aaa", )

        ok_t.write_text(ok_t_content)

        with unittest.mock.patch("builtins.input") as mock_input:
            mock_input.return_value = str(ok_t)
            c_r = TT.render(str(ng_t), ctx, ask=True)
            self.assertEqual(c_r, ok_content)
        try:
            TT.render(str(ng_t), ctx, ask=False)
            assert False  # force raising an exception.
        except TT.jinja2.TemplateNotFound:
            pass

    def test_24_render__wo_paths(self):
        fname = self.templates[0][0]
        workdir = pathlib.Path(self.workdir)

        assert workdir / fname

        subdir = workdir / "a/b/c"
        subdir.mkdir(parents=True)

        tmpl = subdir / fname
        tmpl.write_text("{{ a|default('aaa') }}")

        c_r = TT.render(str(tmpl))
        self.assertEqual(c_r, "aaa")

    def test_25_render__w_prefer_paths(self):
        workdir = pathlib.Path(self.workdir / 'a' / 'b' / 'c')
        workdir.mkdir(parents=True)

        assert workdir.exists()

        tmpl_ref = pathlib.Path(self.workdir / 'ref_25_0.j2')
        tmpl_ref.write_text("{{ a | d('A') }}\n")

        tmpl = pathlib.Path(workdir / 'd.j2')
        tmpl.write_text("{{ a | d('xyz') }}\n")

        c_r = TT.render(tmpl, paths=[self.workdir])
        self.assertNotEqual(c_r, 'A')
        self.assertEqual(c_r, 'xyz')

    def test_30_try_render_with_empty_filepath_and_content(self):
        self.assertRaises(ValueError, TT.try_render)

    def test_40_render__w_filter(self):
        fname, _, ctx = TMPL_WITH_FILTER
        fpath = pathlib.Path(self.workdir) / fname
        c_r = TT.render(str(fpath), filters={"negate": negate})
        self.assertEqual(c_r, ctx)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring, invalid-name
import operator
import unittest

import anyconfig.processors.processors as TT

from .common import A, A2, B, C, PRS


class Test_10_Processor(unittest.TestCase):

    def test_10_eq(self):
        (a1, a2, a22, b) = (A(), A(), A2(), B())
        self.assertEqual(a1, a2)
        self.assertNotEqual(a1, b)
        self.assertNotEqual(a1, a22)


class Test_40_Processors(unittest.TestCase):

    def test_10_init(self):
        prcs = TT.Processors()
        self.assertFalse(prcs.list())

    def test_12_init_with_processors(self):
        prcs = TT.Processors(PRS)
        self.assertEqual(prcs.list(sort=True),
                         sorted(PRS, key=operator.methodcaller('cid')))

    def test_20_list_by_cid(self):
        prcs = TT.Processors(PRS)
        exp = sorted(((p.cid(), [p]) for p in PRS),
                     key=TT.operator.itemgetter(0))
        self.assertEqual(prcs.list_by_cid(), exp)

    def test_20_list_x(self):
        prcs = TT.Processors(PRS)
        self.assertRaises(ValueError, prcs.list_x)
        self.assertEqual(prcs.list_x('cid'),
                         sorted(set(p.cid() for p in PRS)))
        self.assertEqual(prcs.list_x('type'),
                         sorted(set(p.type() for p in PRS)))

        res = sorted(set(A.extensions() + B.extensions() + C.extensions()))
        self.assertEqual(prcs.list_x('extension'), res)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
"""File based test data collector.
"""
import pathlib
import typing

from anyconfig.api import InDataExT


DictT = typing.Dict[str, typing.Any]
MaybePathT = typing.Optional[pathlib.Path]


class TDataPaths(typing.NamedTuple):
    """A namedtuple object keeps test data paths."""
    datadir: pathlib.Path
    inp: pathlib.Path
    exp: MaybePathT
    opts: MaybePathT
    scm: MaybePathT
    query: MaybePathT
    ctx: MaybePathT


class TData(typing.NamedTuple):
    """A namedtuple object keeps test data.
    """
    datadir: pathlib.Path
    inp_path: pathlib.Path
    inp: InDataExT
    exp: DictT
    opts: DictT
    scm: typing.Union[pathlib.Path, str]
    query: str
    ctx: DictT

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Utility functions to list and find appropriate parser class objects and
instances.

.. versionchanged:: 0.10.2

   - Split and re-organize the module and export only some functions.

.. versionadded:: 0.9.5

   - Add to abstract processors such like Parsers (loaders and dumpers).
"""
from .detectors import (
    is_iterable, is_path, is_path_obj, is_ioinfo,
    is_stream_ioinfo, is_path_like_object, is_paths,
    is_dict_like, is_list_like
)
from .files import get_path_from_stream
from .lists import (
    groupby, concat
)
from .utils import (
    filter_options, noop
)


__all__ = [
    'is_iterable', 'is_path', 'is_path_obj', 'is_ioinfo',
    'is_stream_ioinfo', 'is_path_like_object', 'is_paths',
    'is_dict_like', 'is_list_like',
    'get_path_from_stream',
    'groupby', 'concat',
    'filter_options', 'noop',
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, invalid-name
import operator
import unittest

import anyconfig.parsers.utils as TT

from anyconfig.common import (
    UnknownFileTypeError, UnknownProcessorTypeError
)
from anyconfig.backend.json import PARSERS as JSON_PSR_CLSS
from anyconfig.parsers.parsers import Parsers


PSRS = Parsers().list()
JSON_PSRS = sorted(
    (p() for p in JSON_PSR_CLSS),
    key=operator.methodcaller('priority'), reverse=True
)


class TestCase(unittest.TestCase):

    def test_load_plugins(self):
        TT.load_plugins()
        self.assertTrue(PSRS)

    def test_list_types(self):
        res = TT.list_types()
        self.assertTrue(bool(res))
        self.assertTrue(any(x in res for x in ('json', 'ini', 'xml')))

    def test_list_by_x(self):
        for lfn in (TT.list_by_cid, TT.list_by_type, TT.list_by_extension):
            psrs = lfn()
            self.assertTrue(bool(psrs))

    def test_findall_ng_cases(self):
        ies = (((None, None), ValueError),  # w/o path nor type
               (('/tmp/x.xyz', None), UnknownFileTypeError),
               (('/dev/null', None), UnknownFileTypeError),
               ((None, 'xyz'), UnknownProcessorTypeError),
               )
        for inp, exc in ies:
            with self.assertRaises(exc):
                TT.findall(*inp)

    def test_findall(self):
        argss = (('foo.json', None),
                 (None, 'json'),
                 )
        for args in argss:
            psrs = TT.findall(*args)

            self.assertTrue(bool(psrs))
            self.assertEqual(psrs, JSON_PSRS)

    def test_find(self):
        argss = (('foo.json', None),
                 (None, 'json'),
                 (None, JSON_PSR_CLSS[0]),
                 (None, JSON_PSRS[0]),
                 )
        for args in argss:
            psr = TT.find(*args)
            self.assertEqual(psr, JSON_PSRS[0])

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Some common constants, utility functions and so on.
"""
import typing


ResultT = typing.Tuple[bool, typing.Union[str, typing.List[str]]]

# vim:sw=4:ts=4:et:



#
# Forked from m9dicts.tests.{api,dicts}
#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
#
# pylint: disable=missing-docstring,invalid-name
import anyconfig.dicts as TT

from .. import base
from . import common


class TestCase(common.TestCase):
    kind = 'merge'

    def test_merge(self):
        for data in self.each_data():
            upd = base.load_data(data.scm, ordered=True)  # diversion.
            TT.merge(data.inp, upd, **data.opts)
            self.assertEqual(data.inp, data.exp, data)

    def test_merge_with_a_dict(self):
        for data in self.each_data():
            upd = base.load_data(data.scm)
            TT.merge(data.inp, upd, **data.opts)
            self.assertEqual(data.inp, data.exp, data)

    def test_merge_with_an_iterable(self):
        for data in self.each_data():
            upd = base.load_data(data.scm).items()
            TT.merge(data.inp, upd, **data.opts)
            self.assertEqual(data.inp, data.exp, data)

    def test_merge_with_invalid_data(self):
        with self.assertRaises((ValueError, TypeError)):
            TT.merge(dict(a=1), 1)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Functions to detect something.
"""
import collections.abc
import pathlib
import types
import typing

from ..common import (
    IOInfo, IOI_KEYS, IOI_STREAM
)


GLOB_MARKER = '*'


def is_iterable(obj: typing.Any) -> bool:
    """
    >>> is_iterable([])
    True
    >>> is_iterable(())
    True
    >>> is_iterable([x for x in range(10)])
    True
    >>> is_iterable((1, 2, 3))
    True
    >>> g = (x for x in range(10))
    >>> is_iterable(g)
    True
    >>> is_iterable('abc')
    False
    >>> is_iterable(0)
    False
    >>> is_iterable({})
    False
    """
    return isinstance(obj, (list, tuple, types.GeneratorType)) or \
        (not isinstance(obj, (int, str, dict)) and
         bool(getattr(obj, 'next', False)))


def is_path(obj: typing.Any) -> bool:
    """
    Is given object 'obj' a file path?

    :param obj: file path or something
    :return: True if 'obj' is a file path
    """
    return isinstance(obj, str)


def is_path_obj(obj: typing.Any) -> bool:
    """Is given object 'input' a pathlib.Path object?

    :param obj: a pathlib.Path object or something
    :return: True if 'obj' is a pathlib.Path object

    >>> obj = pathlib.Path(__file__)
    >>> assert is_path_obj(obj)
    >>> assert not is_path_obj(__file__)
    """
    return isinstance(obj, pathlib.Path)


def is_file_stream(obj: typing.Any) -> bool:
    """Is given object 'input' a file stream (file/file-like object)?

    :param obj: a file / file-like (stream) object or something
    :return: True if 'obj' is a file stream

    >>> assert is_file_stream(open(__file__))
    >>> assert not is_file_stream(__file__)
    """
    return callable(getattr(obj, 'read', False))


def is_ioinfo(obj: typing.Any) -> bool:
    """
    :return: True if given 'obj' is a 'IOInfo' namedtuple object.
    """
    if isinstance(obj, tuple) and getattr(obj, '_asdict', False):
        return all(k in typing.cast(IOInfo, obj)._asdict() for k in IOI_KEYS)

    return False


def is_stream_ioinfo(obj: typing.Any) -> bool:
    """
    :param obj: IOInfo object or something
    :return: True if given IOInfo object 'obj' is of file / file-like object

    >>> ioi = IOInfo(None, IOI_STREAM, None, None)
    >>> assert is_stream_ioinfo(ioi)
    >>> assert not is_stream_ioinfo(__file__)
    """
    return getattr(obj, 'type', None) == IOI_STREAM


def is_path_like_object(obj: typing.Any, marker: str = GLOB_MARKER) -> bool:
    """
    Is given object 'obj' a path string, a pathlib.Path, a file / file-like
    (stream) or IOInfo namedtuple object?

    :param obj:
        a path string, pathlib.Path object, a file / file-like or 'IOInfo'
        object

    :return:
        True if 'obj' is a path string or a pathlib.Path object or a file
        (stream) object

    >>> assert is_path_like_object(__file__)
    >>> assert not is_path_like_object('/a/b/c/*.json', '*')

    >>> assert is_path_like_object(pathlib.Path('a.ini'))
    >>> assert not is_path_like_object(pathlib.Path('x.ini'), 'x')
    >>> assert is_path_like_object(open(__file__))
    """
    return ((is_path(obj) and marker not in obj) or
            (is_path_obj(obj) and marker not in obj.as_posix()) or
            is_file_stream(obj) or is_ioinfo(obj))


def is_dict_like(obj: typing.Any) -> bool:
    """
    :param obj: Any object behaves like a dict.

    >>> is_dict_like('a string')
    False
    >>> is_dict_like({})
    True
    >>> import collections
    >>> is_dict_like(collections.OrderedDict((('a', 1), ('b', 2))))
    True
    """
    return isinstance(obj, (dict, collections.abc.Mapping))  # any others?


_LIST_LIKE_TYPES = (collections.abc.Iterable, collections.abc.Sequence)


def is_list_like(obj: typing.Any) -> bool:
    """
    >>> is_list_like([])
    True
    >>> is_list_like(())
    True
    >>> is_list_like([x for x in range(10)])
    True
    >>> is_list_like((1, 2, 3))
    True
    >>> g = (x for x in range(10))
    >>> is_list_like(g)
    True
    >>> is_list_like('abc')
    False
    >>> is_list_like(0)
    False
    >>> is_list_like({})
    False
    """
    return isinstance(obj, _LIST_LIKE_TYPES) and \
        not (isinstance(obj, str) or is_dict_like(obj))


def is_paths(maybe_paths: typing.Any, marker: str = '*') -> bool:
    """
    Does given object 'maybe_paths' consist of path or path pattern strings?
    """
    if is_file_stream(maybe_paths) or is_ioinfo(maybe_paths):
        return False

    return (
        (is_path(maybe_paths) and marker in maybe_paths) or  # Path str
        (is_path_obj(maybe_paths) and marker in str(maybe_paths)) or
        (is_list_like(maybe_paths) and
         all(
            is_path(p) or is_ioinfo(p) or is_path_obj(p) or is_file_stream(p)
            for p in maybe_paths)
         )
    )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh @ gmmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=invalid-name
r"""ioinfo.main to provide internal APIs used from other modules.
"""
import pathlib
import typing

from .. import common


def path_and_ext(path: pathlib.Path) -> typing.Tuple[pathlib.Path, str]:
    """Normaliez path objects and retunr it with file extension.
    """
    abs_path = path.expanduser().resolve()
    file_ext = abs_path.suffix
    return (
        abs_path,
        file_ext[1:] if file_ext.startswith('.') else ''
    )


def from_path_object(path: pathlib.Path) -> common.IOInfo:
    """
    Return an IOInfo object made from :class:`pathlib.Path` object ``path``.
    """
    (abs_path, file_ext) = path_and_ext(path)

    return common.IOInfo(
        abs_path, common.IOI_PATH_OBJ, str(abs_path), file_ext
    )


def from_io_stream(strm: typing.IO) -> common.IOInfo:
    """
    Return an IOInfo object made from IO stream object ``strm``.
    """
    path = getattr(strm, 'name', '')
    if path:
        (abs_path, file_ext) = path_and_ext(pathlib.Path(path))
    else:
        (abs_path, file_ext) = (path, '')

    return common.IOInfo(
        strm, common.IOI_STREAM, str(abs_path), file_ext
    )


def make(obj: typing.Any) -> common.IOInfo:
    """
    :param obj: a path string, a pathlib.Path or a file / file-like object
    :return:
        Namedtuple object represents a kind of input object such as a file /
        file-like object, path string or pathlib.Path object

    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    if isinstance(obj, common.IOInfo):
        return obj

    if isinstance(obj, str):
        obj = pathlib.Path(obj)

    if isinstance(obj, pathlib.Path):
        return from_path_object(obj)

    # Which is better? isinstance(obj, io.IOBase):
    if getattr(obj, 'read', False):
        return from_io_stream(obj)

    raise ValueError(repr(obj))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
"""Common data types for api.multi_load test cases.
"""
import pathlib
import typing


DictT = typing.Dict[str, typing.Any]


class TData(typing.NamedTuple):
    """A namedtuple object keeps test data.
    """
    datadir: pathlib.Path
    inputs: typing.List[pathlib.Path]  # Same as the above.
    exp: DictT
    opts: DictT
    scm: pathlib.Path
    query: str
    ctx: DictT

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2017 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=bare-except
#
r"""anyconfig.query module to support query data with JMESPath expressions.

Changelog:

.. versionadded:: 0.8.3

   - Added to query config data with JMESPath expression, http://jmespath.org
"""
import typing
import warnings

import jmespath

from ..common import (
    InDataExT, InDataT
)
from ..utils import is_dict_like
from .datatypes import MaybeJexp


def try_query(data: InDataExT, jexp: MaybeJexp = None, **options) -> InDataExT:
    """
    Try to query data with JMESPath expression `jexp`.
    """
    if jexp is None or not jexp:
        return data

    if not is_dict_like(data):  # Some primitive types like int, str.
        warnings.warn('Could not query because given data is not '
                      f'a mapping object (type? {type(data)}')
        return data

    (odata, exc) = query(
        typing.cast(InDataT, data), typing.cast(str, jexp), **options
    )
    if exc:
        raise exc

    return odata  # type: ignore


def query(data: InDataT, jexp: str, **_options
          ) -> typing.Tuple[typing.Optional[InDataT],
                            typing.Optional[Exception]]:
    """
    Filter data with given JMESPath expression.

    See also: https://github.com/jmespath/jmespath.py and http://jmespath.org.

    :param data: Target object (a dict or a dict-like object) to query
    :param jexp: a string represents JMESPath expression
    :param options: Keyword options

    :return: A tuple of query result and maybe exception if failed
    """
    exc: typing.Optional[Exception] = None
    try:
        pexp = jmespath.compile(jexp)
        return (pexp.search(data), exc)

    except ValueError as exc:  # jmespath.exceptions.*Error inherit from it.
        return (data, exc)

    except:  # noqa: E722
        return (None, exc)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

import anyconfig.query

from . import common


@unittest.skipIf(not anyconfig.query.SUPPORTED,
                 'jmespath lib is not available')
class TestCase(common.TestCase):
    kind = 'query'

    def test_single_load(self):
        for data in self.each_data():
            self.assertEqual(
                self.target_fn(
                    data.inp_path, ac_query=data.query.strip(), **data.opts
                ),
                data.exp,
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_single_load_with_invalid_query_string(self):
        for data in self.each_data():
            self.assertEqual(
                self.target_fn(
                    data.inp_path, ac_query=None, **data.opts
                ),
                self.target_fn(data.inp_path, **data.opts),
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_single_load_intentional_failures(self):
        for data in self.each_data():
            with self.assertRaises(AssertionError):
                exp = dict(z=1, zz='zz', zzz=[1, 2, 3], zzzz=dict(z=0))
                self.assertEqual(
                    self.target_fn(
                        data.inp_path, ac_query=data.query, **data.opts
                    ),
                    exp
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Backend basic classes, functions and constants.
"""
import typing

from .compat import BinaryFilesMixin
from .dumpers import (
    ToStringDumperMixin, ToStreamDumperMixin, BinaryDumperMixin
)
from .loaders import (
    FromStringLoaderMixin, FromStreamLoaderMixin, BinaryLoaderMixin
)
from .utils import (
    ensure_outdir_exists, to_method
)
from .parsers import (
    Parser,
    StringParser, StreamParser, StringStreamFnParser,
)


ParserT = typing.TypeVar('ParserT', bound=Parser)
ParsersT = typing.List[ParserT]
ParserClssT = typing.List[typing.Type[ParserT]]


__all__ = [
    'BinaryFilesMixin',
    'ToStringDumperMixin', 'ToStreamDumperMixin', 'BinaryDumperMixin',
    'FromStringLoaderMixin', 'FromStreamLoaderMixin', 'BinaryLoaderMixin',
    'ensure_outdir_exists', 'to_method',
    'Parser',
    'StringParser', 'StreamParser', 'StringStreamFnParser',
    'ParserT', 'ParsersT', 'ParserClssT',
]

# vim:sw=4:ts=4:et:



#
# Forked from m9dicts.tests.{api,dicts}
#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
#
# pylint: disable=missing-docstring,invalid-name
import anyconfig.dicts as TT

from .. import base
from . import common


class TestCase(common.TestCase):
    kind = 'get'

    def test_get(self):
        for data in self.each_data():
            emsg = base.load_data(data.scm)  # diversion.
            (res, err) = TT.get(data.inp, data.query)

            if emsg:
                self.assertTrue(bool(err), data)
            else:  # emsg = ''
                self.assertEqual(err, '', data)

            self.assertEqual(res, data.exp, data)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, relative-beyond-top-level
from ..multi_load import test_basics as multi
from ..single_load import test_basics as single

from . import common


class SingleTestCase(common.SingleBase, single.TestCase):
    pass


class MultiTestCase(common.MultiBase, multi.TestCase):
    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Utility funtions to process file and file paths.
"""
import pathlib
import typing

from ..common import (
    IOInfo, PathOrIOInfoT
)
from . import factory, utils


PathOrPathsT = typing.Union[
    typing.Iterable[PathOrIOInfoT],
    PathOrIOInfoT
]


def expand_paths_itr(paths: PathOrPathsT) -> typing.Iterator[PathOrIOInfoT]:
    """
    :param paths:
        A glob path pattern string or pathlib.Path object holding such path, or
        a list consists of path strings or glob path pattern strings or
        pathlib.Path object holding such ones.
    """
    if isinstance(paths, IOInfo):
        yield typing.cast(IOInfo, paths)

    elif isinstance(paths, (str, pathlib.Path)):
        if isinstance(paths, pathlib.Path):
            paths = str(paths)

        (base, pattern) = utils.split_path_by_marker(paths)

        if not pattern:
            yield pathlib.Path(base)
            return

        base_2 = pathlib.Path(base or '.').resolve()
        for path in sorted(base_2.glob(pattern)):
            yield path

    elif utils.is_io_stream(paths):
        yield paths  # type: ignore

    else:
        for path in paths:  # type: ignore
            for cpath in expand_paths_itr(path):
                yield cpath


def expand_paths(paths: PathOrPathsT) -> typing.List[IOInfo]:
    """
    :param paths:
        A glob path pattern string or pathlib.Path object holding such path, or
        a list consists of path strings or glob path pattern strings or
        pathlib.Path object holding such ones, or file objects
    """
    return [factory.make(p) for p in expand_paths_itr(paths)]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Common utilities for test cases.
"""
import os
import pathlib
import shutil
import tempfile
import unittest


CNF_0 = dict(name="a", a=1, b=dict(b=[0, 1], c="C"))
SCM_0 = {"type": "object",
         "properties": {
             "name": {"type": "string"},
             "a": {"type": "integer"},
             "b": {"type": "object",
                   "properties": {
                       "b": {"type": "array",
                             "items": {"type": "integer"}}}}}}
# :seealso: tests/res/00-cnf.json
CNF_1 = {"a": 1, "b": {"b": [1, 2], "c": "C"}, "name": "aaa"}


def selfdir():
    """
    >>> selfdir().exists()
    True
    """
    return pathlib.Path(__file__).parent.resolve()


def respath(filename: str) -> str:
    """
    :pattern: Resource path path or path glob pattern

    >>> assert respath("00-cnf.json").endswith('/res/00-cnf.json')
    >>>
    """
    return str(selfdir() / "res" / filename)


def setup_workdir():
    """
    >>> workdir = setup_workdir()
    >>> assert workdir != '.'
    >>> assert workdir != '/'
    >>> pathlib.Path(workdir).exists()
    True
    >>> os.rmdir(workdir)
    """
    return tempfile.mkdtemp(dir="/tmp", prefix="python-anyconfig-tests-")


def cleanup_workdir(workdir):
    """
    FIXME: Danger!

    >>> from os import linesep as lsep
    >>> workdir = pathlib.Path(setup_workdir())
    >>> workdir.exists()
    True
    >>> _ = (workdir / "workdir.stamp").open('w').write("OK!" + lsep)
    >>> cleanup_workdir(str(workdir))
    >>> workdir.exists()
    False
    """
    assert workdir != '/'
    assert workdir != '.'

    os.system("rm -rf " + workdir)


class TestCaseWithWorkdir(unittest.TestCase):
    curdir = str(pathlib.Path('.').resolve())

    def setUp(self):
        self.workdir = pathlib.Path(setup_workdir()).resolve()

    def tearDown(self):
        workdir = self.workdir
        ng_dirs = ('/', str(self.workdir.home()), self.curdir)

        if str(workdir) not in ng_dirs:
            shutil.rmtree(workdir)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods

import anyconfig.backend.ini as TT
import tests.backend.common as TBC


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf_s = TBC.read_from_res("20-00-cnf.ini")


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):

    load_options = dict(allow_no_value=False, defaults=None)

    def test_42_loads_invalid_input(self):
        invalid_cnf_s = "key=name"  # No section.
        self.assertRaises(Exception, self.psr.loads, invalid_cnf_s)

    def test_44_loads_with_ac_parse_value_option(self):
        cnf = self.psr.loads(self.cnf_s, ac_parse_value=True)
        ref = self.psr.loads(self.cnf_s)
        ref["DEFAULT"]["a"] = ref["sect0"]["a"] = 0
        ref["DEFAULT"]["c"] = ref["sect0"]["c"] = 5
        ref["sect0"]["d"] = ref["sect0"]["d"].split(',')
        self._assert_dicts_equal(cnf, ref=ref)


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):

    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring, invalid-name
import anyconfig.models.processor


class A(anyconfig.models.processor.Processor):
    _cid = 'A'
    _type = 'json'
    _extensions = ['json', 'jsn', 'js']


class A2(A):
    _cid = 'A2'
    _priority = 20  # Higher priority than A.


class A3(A):
    _cid = 'A3'
    _priority = 99  # Higher priority than A and A2.


class B(anyconfig.models.processor.Processor):
    _cid = 'B'
    _type = 'yaml'
    _extensions = ['yaml', 'yml']
    _priority = 99  # Higher priority than C.


class C(anyconfig.models.processor.Processor):
    _cid = 'dummy'
    _type = 'yaml'
    _extensions = ['yaml', 'yml']


PRS = [A, A2, A3, B, C]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Common utility functions.
"""
import unittest

import anyconfig.api._load as TT

from . import collector


class TestCase(unittest.TestCase, collector.DataCollector):

    @staticmethod
    def target_fn(*args, **kwargs):
        return TT.multi_load(*args, **kwargs)

    def setUp(self):
        self.init()

    def test_multi_load(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn(tdata.inputs, **tdata.opts),
                tdata.exp,
                tdata
            )

    def test_multi_load_failures(self):
        for tdata in self.each_data():
            with self.assertRaises(AssertionError):
                self.assertEqual(
                    self.target_fn(tdata.inputs, **tdata.opts),
                    None,
                    tdata
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# Suppress import positions after some global variables are defined
# pylint: disable=wrong-import-position
"""A module to aggregate config parser (loader/dumper) backends.
"""
import typing

from ..backend import ParserClssT, PARSERS
from ..processors import Processors
from ..singleton import Singleton


class Parsers(Processors, Singleton):
    """
    Manager class for parsers.
    """
    _pgroup: str = 'anyconfig_backends'

    def __init__(self, prcs: typing.Optional[ParserClssT] = None
                 ) -> None:
        """Initialize with PARSERS.
        """
        if prcs is None:
            prcs = PARSERS

        super().__init__(prcs)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2018 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods
# pylint: disable=ungrouped-imports

try:
    import anyconfig.backend.json.simplejson as TT
except ImportError:
    import unittest
    raise unittest.SkipTest

from .default import TBC, CNF_0_S, CNF_0


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf_s = CNF_0_S
    cnf = CNF_0


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):

    load_options = dump_options = dict(parse_int=None, indent=2)
    empty_patterns = [('', {}), ('{}', {}), ('[]', []), ('null', None)]

    load_options["use_decimal"] = True


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):
    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, invalid-name
import tempfile
import unittest

import anyconfig.backend.base.dumpers as TT


class DumperMixinTestCase(unittest.TestCase):

    def test_ropen(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            with TT.DumperMixin().wopen(temp_dir + '/test.txt') as fio:
                self.assertEqual(fio.mode, 'w')


class BinaryDumperMixinTestCase(unittest.TestCase):

    def test_ropen(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            with TT.BinaryDumperMixin().wopen(temp_dir + '/test.txt') as fio:
                self.assertEqual(fio.mode, 'wb')

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Common utility functions.
"""
import pathlib
import typing


TESTS_DIR = pathlib.Path(__file__).parent.parent.resolve()
RES_DIR = TESTS_DIR / 'res'

NULL_CNTNR: typing.Dict[str, typing.Any] = dict()

DATA = '3.149265'  # for test_utils.py

# vim:sw=4:ts=4:et:



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods
# pylint: disable=ungrouped-imports,protected-access

import io
import unittest

import anyconfig.backend.xml as TT
import tests.backend.common as TBC

from tests.backend.common import to_bytes


CNF_0 = {'config': {'@attrs': {'val:name': 'foo',
                               'xmlns': 'http://example.com/ns/cnf',
                               'xmlns:val': 'http://example.com/ns/cnf/val'},
                    'val:a': '0',
                    'val:b': {'@attrs': {'id': 'b0'}, '@text': 'bbb'},
                    'val:c': None,
                    'sect0': {'val:d': 'x, y, z'},
                    'list1': [{'item': '0'}, {'item': '1'}, {'item': '2'}],
                    'list2': {'@attrs': {'id': 'list2'},
                              '@children': [{'item': 'i'},
                                            {'item': 'j'}]}}}


class Test_00(unittest.TestCase):

    def test_10__namespaces_from_file(self):
        ref = {"http://example.com/ns/config": '',
               "http://example.com/ns/config/val": "val"}
        xmlfile = io.StringIO(TBC.read_from_res("20-00-cnf.xml"))
        self.assertEqual(TT._namespaces_from_file(xmlfile), ref)

    def test_20__process_elem_text__whitespaces(self):
        (elem, dic, subdic) = (TT.ET.XML("<a> </a>"), {}, {})
        TT._process_elem_text(elem, dic, subdic)
        self.assertTrue(not dic)
        self.assertTrue(not subdic)

    def test_22__process_elem_text__wo_attrs_and_children(self):
        (elem, dic, subdic) = (TT.ET.XML("<a>A</a>"), {}, {})
        TT._process_elem_text(elem, dic, subdic, text="#text")
        self.assertEqual(dic, {"a": 'A'})
        self.assertTrue(not subdic)

    def test_22__process_elem_text__wo_attrs_and_children_parse(self):
        (elem, dic, subdic) = (TT.ET.XML("<a>A</a>"), {}, {})
        TT._process_elem_text(elem, dic, subdic, text="#text",
                              ac_parse_value=True)
        self.assertEqual(dic, {"a": 'A'})
        self.assertTrue(not subdic)

        (elem, dic, subdic) = (TT.ET.XML("<a>1</a>"), {}, {})
        TT._process_elem_text(elem, dic, subdic, text="#text",
                              ac_parse_value=True)
        self.assertEqual(dic, {"a": 1})
        self.assertTrue(not subdic)

    def test_24__process_elem_text__w_attrs(self):
        (elem, dic, subdic) = (TT.ET.XML("<a id='1'>A</a>"), {}, {})
        TT._process_elem_text(elem, dic, subdic, text="#text")
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"#text": 'A'})

    def test_24__process_elem_text__w_children(self):
        (elem, dic, subdic) = (TT.ET.XML("<a>A<b/></a>"), {}, {})
        TT._process_elem_text(elem, dic, subdic, text="#text")
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"#text": 'A'})

    def test_30__process_elem_attrs__wo_text_and_children(self):
        (elem, dic, subdic) = (TT.ET.XML("<a id='A'/>"), {}, {})
        TT._process_elem_attrs(elem, dic, subdic)
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"@attrs": {"id": 'A'}})

    def test_32__process_elem_attrs__w_text(self):
        (elem, dic, subdic) = (TT.ET.XML("<a id='A'>AAA</a>"), {}, {})
        TT._process_elem_attrs(elem, dic, subdic)
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"@attrs": {"id": 'A'}})

    def test_34__process_elem_attrs__merge_attrs(self):
        (elem, dic, subdic) = (TT.ET.XML("<a id='A'/>"), {}, {})
        TT._process_elem_attrs(elem, dic, subdic, merge_attrs=True)
        self.assertEqual(dic, {"a": {"id": 'A'}})
        self.assertTrue(not subdic)

    def test_36__process_elem_attrs__wo_text_and_children_parse(self):
        (elem, dic, subdic) = (TT.ET.XML("<a id='1'/>"), {}, {})
        TT._process_elem_attrs(elem, dic, subdic, ac_parse_value=True)
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"@attrs": {"id": 1}})

        (elem, dic, subdic) = (TT.ET.XML("<a id='A'/>"), {}, {})
        TT._process_elem_attrs(elem, dic, subdic, ac_parse_value=True)
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"@attrs": {"id": 'A'}})

        (elem, dic, subdic) = (TT.ET.XML("<a id='true'/>"), {}, {})
        TT._process_elem_attrs(elem, dic, subdic, ac_parse_value=True)
        self.assertTrue(not dic)
        self.assertEqual(subdic, {"@attrs": {"id": True}})

    def test_40__process_children_elems__root(self):
        (elem, dic, subdic) = (TT.ET.XML("<list><i>A</i><i>B</i></list>"), {},
                               {})
        TT._process_children_elems(elem, dic, subdic)
        self.assertEqual(dic, {"list": [{"i": "A"}, {"i": "B"}]})
        self.assertTrue(not subdic)

    def test_42__process_children_elems__w_attr(self):
        (elem, dic) = (TT.ET.XML("<list id='xyz'><i>A</i><i>B</i></list>"), {})
        subdic = {"id": "xyz"}
        ref = subdic.copy()
        ref.update({"#children": [{"i": "A"}, {"i": "B"}]})

        TT._process_children_elems(elem, dic, subdic, children="#children")
        self.assertTrue(not dic)
        self.assertEqual(subdic, ref, subdic)

    def test_44__process_children_elems__w_children_have_unique_keys(self):
        (elem, dic, subdic) = (TT.ET.XML("<a><x>X</x><y>Y</y></a>"), {}, {})
        TT._process_children_elems(elem, dic, subdic)
        self.assertEqual(dic, {"a": {"x": "X", "y": "Y"}})
        self.assertTrue(not subdic)

    def test_46__process_children_elems__w_merge_attrs(self):
        elem = TT.ET.XML("<a z='Z'><x>X</x><y>Y</y></a>")
        dic = {"a": {"@attrs": {"z": "Z"}}}
        subdic = dic["a"]["@attrs"]
        TT._process_children_elems(elem, dic, subdic, merge_attrs=True)
        self.assertEqual(dic, {"a": {"x": "X", "y": "Y", "z": "Z"}}, dic)


class Test_00_1(unittest.TestCase):

    def _assert_eq_dic_from_snippet(self, snippet, ref, **opts):
        self.assertEqual(TT.elem_to_container(TT.ET.XML(snippet), **opts), ref)

    def test_10_elem_to_container__None(self):
        self.assertEqual(TT.elem_to_container(None), dict())

    def test_10_root_to_container__None(self):
        self.assertEqual(TT.root_to_container(None), dict())

    def test_12_elem_to_container__empty(self):
        self._assert_eq_dic_from_snippet("<a/>", dict(a=None))

    def test_20_elem_to_container__attrs(self):
        ref = dict(a={"@attrs": dict(x='1', y='y')})
        self._assert_eq_dic_from_snippet("<a x='1' y='y'/>", ref)

    def test_30_elem_to_container__child(self):
        ref = dict(a=dict(b="b"))
        self._assert_eq_dic_from_snippet("<a><b>b</b></a>", ref)

    def test_32_elem_to_container__children__same_keys(self):
        ref = {'a': [{'b': '1'}, {'b': '2'}]}
        self._assert_eq_dic_from_snippet("<a><b>1</b><b>2</b></a>", ref)

    def test_34_elem_to_container__children(self):
        ref = {'a': {'b': 'b', 'c': 'c'}}
        self._assert_eq_dic_from_snippet("<a><b>b</b><c>c</c></a>", ref)

    def test_36_elem_to_container__children__same_keys_w_text(self):
        ref = {'a': {'@text': 'aaa', '@children': [{'b': '1'}, {'b': '2'}]}}
        self._assert_eq_dic_from_snippet("<a>aaa<b>1</b><b>2</b></a>", ref)

    def test_40_elem_to_container__text(self):
        self._assert_eq_dic_from_snippet("<a>A</a>", {'a': 'A'})

    def test_42_elem_to_container__text_attrs(self):
        ref = dict(a={"@attrs": {'x': 'X'}, "@text": "A"})
        self._assert_eq_dic_from_snippet("<a x='X'>A</a>", ref)

    def test_50_root_to_container__text_attrs_tags(self):
        ref = dict(a={"_attrs": {'x': 'X'}, "_text": "A"})
        tags = dict(attrs="_attrs", text="_text")
        self.assertEqual(TT.root_to_container(TT.ET.XML("<a x='X'>A</a>"),
                                              dict, {}, tags=tags),
                         ref)


def tree_to_string(tree):
    return TT.ET.tostring(tree.getroot())


class Test_00_2(unittest.TestCase):

    def test_00_container_to_etree__None(self):
        self.assertTrue(TT.container_to_etree(None) is None)

    def test_10_container_to_etree__text_attrs(self):
        ref = to_bytes('<a x="X" y="Y">A</a>')
        obj = dict(a={"@attrs": {'x': 'X', 'y': 'Y'}, "@text": "A"})
        res = TT.container_to_etree(obj)
        self.assertEqual(tree_to_string(res), ref)

    def test_12_container_to_etree__text_attrs_tags(self):
        ref = to_bytes('<a x="X" y="Y">A</a>')
        obj = dict(a={"_attrs": {'x': 'X', 'y': 'Y'}, "_text": "A"})
        tags = dict(attrs="_attrs", text="_text")
        res = TT.container_to_etree(obj, tags=tags)
        self.assertEqual(tree_to_string(res), ref)

    def test_20_container_to_etree__child(self):
        ref = to_bytes("<a><b>b</b></a>")
        obj = dict(a=dict(b="b"))
        res = TT.container_to_etree(obj)
        self.assertEqual(tree_to_string(res), ref)

    def test_22_container_to_etree__children(self):
        ref = to_bytes("<a><b>b</b><c>c</c></a>")
        obj = {'a': {'@children': [{'b': 'b'}, {'c': 'c'}]}}
        res = TT.container_to_etree(obj)
        self.assertEqual(tree_to_string(res), ref)


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf = CNF_0
    cnf_s = to_bytes(TBC.read_from_res("20-10-cnf.xml"))


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):

    load_options = dump_options = dict(ac_parse_value=False)


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):

    def test_40_load_w_options(self):
        cnf = self.psr.load(self.ioi, ac_parse_value=False)
        self._assert_dicts_equal(cnf)

    def test_42_dump_with_special_option(self):
        ioi = self._to_ioinfo(self.cnf_path)
        self.psr.dump(self.cnf, ioi, ac_parse_value=False)
        cnf = self.psr.load(self.ioi)
        self._assert_dicts_equal(cnf)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Misc utility routines for anyconfig module.
"""
import itertools
import typing


def groupby(itr: typing.Iterable[typing.Any],
            key_fn: typing.Optional[typing.Callable[..., typing.Any]] = None
            ) -> typing.Iterable[
                typing.Tuple[typing.Any, typing.Iterable[typing.Any]]
            ]:
    """
    An wrapper function around itertools.groupby to sort each results.

    :param itr: Iterable object, a list/tuple/genrator, etc.
    :param key_fn: Key function to sort 'itr'.
    """
    return itertools.groupby(sorted(itr, key=key_fn), key=key_fn)


def concat(xss: typing.Iterable[typing.Iterable[typing.Any]]
           ) -> typing.List[typing.Any]:
    """
    Concatenates a list of lists.
    """
    return list(itertools.chain.from_iterable(xs for xs in xss))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Test cases for anyconfig.utils.files.
"""
import pathlib
import unittest

import anyconfig.api.utils as TT

from anyconfig.ioinfo import make as ioinfo_make


class TestCase(unittest.TestCase):

    def test_are_same_file_types(self):
        fun = TT.are_same_file_types
        this_py = pathlib.Path(__file__)
        this = ioinfo_make(this_py)
        other = ioinfo_make(this_py.parent / 'setup.cfg')

        for inp, exp in (([], False),
                         ([this], True),
                         ([this, this], True),
                         ([this, other], False),
                         ([this, other], False),
                         ):
            (self.assertTrue if exp else self.assertFalse)(fun(inp))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-import,import-error,invalid-name
r"""Public APIs to dump configurations data.
"""
from ..common import PathOrIOInfoT, InDataExT
from ..ioinfo import make as ioinfo_make
from ..parsers import find, MaybeParserT
from .datatypes import ParserT


def dump(data: InDataExT, out: PathOrIOInfoT,
         ac_parser: MaybeParserT = None, **options) -> None:
    """
    Save 'data' to 'out'.

    :param data: A mapping object may have configurations data to dump
    :param out:
        An output file path, a file, a file-like object, :class:`pathlib.Path`
        object represents the file or a namedtuple 'anyconfig.common.IOInfo'
        object represents output to dump some data to.
    :param ac_parser: Forced parser type or parser object
    :param options:
        Backend specific optional arguments, e.g. {"indent": 2} for JSON
        loader/dumper backend

    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    ioi = ioinfo_make(out)
    psr: ParserT = find(ioi, forced_type=ac_parser)
    psr.dump(data, ioi, **options)


def dumps(data: InDataExT, ac_parser: MaybeParserT = None,
          **options) -> str:
    """
    Return string representation of 'data' in forced type format.

    :param data: Config data object to dump
    :param ac_parser: Forced parser type or ID or parser object
    :param options: see :func:`dump`

    :return: Backend-specific string representation for the given data
    :raises: ValueError, UnknownProcessorTypeError
    """
    psr: ParserT = find(None, forced_type=ac_parser)
    return psr.dumps(data, **options)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import json
import pathlib
import unittest
import tempfile

from ... import base
from . import utils as TT


RES_DIR = base.RES_DIR / 'multi_load'
SELF = pathlib.Path(__file__)


class TestCase(unittest.TestCase):

    datadir = RES_DIR / 'basics'

    def test_load_data_or_path_no_data(self):
        with tempfile.TemporaryDirectory() as tdir:
            tmp_path = pathlib.Path(tdir)
            aes = [
                ((tmp_path, ), {}, None),
                ((tmp_path, ), dict(default='abc'), 'abc'),
            ]
            for args, kwargs, exp in aes:
                self.assertEqual(
                    TT.load_data_or_path(*args, **kwargs),
                    exp
                )

    def test_load_data_or_path(self):
        # tests/res/multi_load/basics/00/00.json
        ddir = self.datadir / '00'
        inp = ddir / '00.json'
        aes = [
            # datadir, should_exist, load, default
            # .. seealso:: tests/res/multi_load/basics/00/00.json
            ((ddir, ), dict(load=False), inp),
            ((ddir, ), {}, json.load(inp.open())),
        ]
        for args, kwargs, exp in aes:
            self.assertEqual(
                TT.load_data_or_path(*args, **kwargs),
                exp
            )

    def test_load_data_or_path_failures(self):
        datadir = pathlib.Path().cwd() / 'dir_not_exist'
        aes = [
            ((datadir, ), dict(should_exist=(datadir.name, )), OSError),
        ]
        for args, kwargs, exc in aes:
            with self.assertRaises(exc):
                TT.load_data_or_path(*args, **kwargs)

    def test_each_data_from_dir(self):
        inp_refs = list(self.datadir.glob('*/*.json'))
        assert bool(inp_refs)

        e_refs = list(self.datadir.glob('*/e/*.json'))
        assert bool(e_refs)

        o_refs = list(self.datadir.glob('*/o/*.json'))
        assert bool(o_refs)

        aes = [
            ((self.datadir, ), {}, (1, inp_refs, e_refs, o_refs))
        ]
        for args, kwargs, exp in aes:
            res = list(TT.each_data_from_dir(*args, **kwargs))
            self.assertTrue(bool(res))
            self.assertTrue(len(res) > exp[0])

            for inp in exp[1]:
                self.assertTrue(any(inp in td.inputs for td in res))

            for e_file in exp[2]:
                e_ref = json.load(e_file.open())
                self.assertTrue(any(e_ref == td.exp for td in res))

            for o_file in exp[3]:
                o_ref = json.load(o_file.open())
                self.assertTrue(any(o_ref == td.opts for td in res))

    def test_each_data_from_dir_failures(self):
        with self.assertRaises(ValueError):
            _ = list(TT.each_data_from_dir(SELF))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2017 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Pickle backend:

- Format to support: Pickle
- Requirements: It should be available always.

  - pickle/cPickle in python 2 standard library:
    https://docs.python.org/2/library/pickle.html

  - pickle in python 3 standard library:
    https://docs.python.org/3/library/pickle.html

- Development Status :: 4 - Beta
- Limitations: The parser cannot load some primitive data such like '' (empty
  string), ' ' (white space) and [] (empty list) as these are because of the
  implementation of :func:`anyconfig.backend.base.load_with_fn`.
- Special options: All options of pickle.{load{s,},dump{s,}} should work.

Changelog:

.. versionchanged:: 0.9.7

   - Add support of loading primitives other than mapping objects.

.. versionadded:: 0.8.3
"""
import pickle

from . import base


LOAD_OPTS = ['fix_imports', 'encoding', 'errors']
DUMP_OPTS = ['protocol', 'fix_imports']


class Parser(base.StringStreamFnParser,
             base.BinaryLoaderMixin, base.BinaryDumperMixin):
    """
    Parser for Pickle files.
    """
    _cid = 'pickle'
    _type = 'pickle'
    _extensions = ['pkl', 'pickle']
    _load_opts = LOAD_OPTS
    _dump_opts = DUMP_OPTS
    _allow_primitives = True

    _load_from_string_fn = base.to_method(pickle.loads)
    _load_from_stream_fn = base.to_method(pickle.load)
    _dump_to_string_fn = base.to_method(pickle.dumps)
    _dump_to_stream_fn = base.to_method(pickle.dump)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import collections
import pathlib

import anyconfig.api._load as TT
import anyconfig.parsers

from anyconfig.api import (
    UnknownFileTypeError, UnknownProcessorTypeError
)
from ... import base
from . import common


JSON_PARSER = anyconfig.parsers.find(None, 'json')


class MyDict(collections.OrderedDict):
    """My original dict class keep key orders."""


class TestCase(common.TestCase):

    def test_single_load_from_stream(self):
        for data in self.each_data():
            self.assertEqual(
                TT.single_load(data.inp_path.open(), **data.opts),
                data.exp
            )

    def test_single_load_from_path_str(self):
        for data in self.each_data():
            self.assertEqual(
                TT.single_load(str(data.inp_path), **data.opts),
                data.exp
            )

    def test_single_load_with_ac_parser_by_instance(self):
        for data in self.each_data():
            self.assertEqual(
                TT.single_load(data.inp_path, ac_parser=JSON_PARSER),
                data.exp
            )

    def test_single_load_with_ac_parser_by_id(self):
        for data in self.each_data():
            self.assertEqual(
                TT.single_load(data.inp_path, ac_parser=JSON_PARSER.cid()),
                data.exp
            )

    def test_single_load_with_ac_ordered(self):
        for data in self.each_data():
            self.assertEqual(
                TT.single_load(data.inp_path, ac_ordered=True),
                collections.OrderedDict(data.exp)
            )

    def test_single_load_with_ac_dict(self):
        for data in self.each_data():
            res = TT.single_load(data.inp_path, ac_dict=MyDict)
            self.assertTrue(isinstance(res, MyDict))
            self.assertEqual(res, MyDict(**data.exp))

    def test_single_load_missing_file_failures(self):
        with self.assertRaises(FileNotFoundError):
            TT.single_load('not_exist.json')

    def test_single_load_unknown_file_type_failures(self):
        with self.assertRaises(UnknownFileTypeError):
            TT.single_load('dummy.txt')

    def test_single_load_invalid_parser_object_failures(self):
        with self.assertRaises(ValueError):
            TT.single_load('dummy.txt', ac_parser=object())

    def test_single_load_unknown_processor_type_failures(self):
        for data in self.each_data():
            with self.assertRaises(UnknownProcessorTypeError):
                TT.single_load(
                    data.inp_path, ac_parser='proc_does_not_exist'
                )

    def test_single_load_ignore_missing(self):
        inp = pathlib.Path() / 'conf_file_not_exist.json'
        assert not inp.exists()

        res = TT.single_load(inp, ac_parser='json', ac_ignore_missing=True)
        self.assertEqual(res, base.NULL_CNTNR)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

import anyconfig.api._load as TT
import anyconfig.query

from . import common


@unittest.skipIf(not anyconfig.query.SUPPORTED,
                 'jmespath lib is not available')
class TestCase(common.TestCase):
    kind = 'query'

    def test_loads_with_query(self):
        for data in self.each_data():
            self.assertEqual(
                TT.loads(data.inp, ac_query=data.query, **data.opts),
                data.exp,
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_loads_with_invalid_query(self):
        opts = dict(ac_parser='json')
        for data in self.each_data():
            self.assertEqual(
                TT.loads(data.inp, ac_query=None, **opts),
                TT.single_load(data.inp_path, **opts),
                f'{data.datadir!s}, {data.inp_path!s}'
            )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""A Public API to open configuration files by detecting type automatically.
"""
import typing
import warnings

from ..common import PathOrIOInfoT, IOI_STREAM
from ..ioinfo import make as ioinfo_make
from ..parsers import find, MaybeParserT
from .datatypes import ParserT


# pylint: disable=redefined-builtin
def open(path: PathOrIOInfoT, mode: typing.Optional[str] = None,
         ac_parser: MaybeParserT = None, **options) -> typing.IO:
    """
    Open given configuration file with appropriate open flag.

    :param path: Configuration file path
    :param mode:
        Can be 'r' and 'rb' for reading (default) or 'w', 'wb' for writing.
        Please note that even if you specify 'r' or 'w', it will be changed to
        'rb' or 'wb' if selected backend, xml and configobj for example, for
        given config file prefer that.
    :param options:
        Optional keyword arguments passed to the internal file opening APIs of
        each backends such like 'buffering' optional parameter passed to
        builtin 'open' function.

    :return: A file object or None on any errors
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    if not path:
        raise ValueError(f'Invalid argument, path: {path!r}')

    ioi = ioinfo_make(path)
    if ioi.type == IOI_STREAM:
        warnings.warn(f'Looks already opened stream: {ioi!r}')
        return typing.cast(typing.IO, ioi.src)

    psr: ParserT = find(ioi, forced_type=ac_parser)

    if mode is not None and mode.startswith('w'):
        return psr.wopen(ioi.path, **options)

    return psr.ropen(ioi.path, **options)

# vim:sw=4:ts=4:et:



#
# Forked from m9dicts.tests.{api,dicts}
#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
#
# pylint: disable=missing-docstring,invalid-name
import collections
import unittest

import anyconfig.dicts as TT


class TestCase(unittest.TestCase):

    # FIXME: Add some more test cases
    def test_set_(self):
        aes = (
            ((dict(a=1, b=dict(c=2, )), 'a.b.d', 3),
             dict(a=dict(b=dict(d=3)), b=dict(c=2))),
        )
        for args, exp in aes:
            TT.set_(*args)
            self.assertEqual(args[0], exp)

    # FIXME: Add some more test cases
    def test_convert_to(self):
        OD = collections.OrderedDict
        aes = (
            ((OD((('a', 1), )), False, dict), dict(a=1)),
            ((OD((('a', OD((('b', OD((('c', 1), ))), ))), )), False, dict),
             dict(a=dict(b=dict(c=1)))),
        )
        for args, exp in aes:
            self.assertEqual(TT.convert_to(*args), exp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Common data types for anyconfig.query.
"""
import typing


MaybeJexp = typing.Optional[typing.Union[str, bool]]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, invalid-name
import pathlib
import unittest

import anyconfig.backend.base.parsers as TT
import anyconfig.ioinfo


MZERO = TT.Parser()._container_factory()()


class TestCase(unittest.TestCase):

    def setUp(self):
        self.psr = TT.Parser()

    def test_10_type(self):
        self.assertEqual(self.psr.type(), str(TT.Parser._type))

    def test_20_loads__null_content(self):
        cnf = self.psr.loads('')
        self.assertEqual(cnf, MZERO)
        self.assertTrue(isinstance(cnf, type(MZERO)))

    def test_30_load__ac_ignore_missing(self):
        cpath = pathlib.Path.cwd() / 'conf_file_not_exist.json'
        assert not cpath.exists()

        ioi = anyconfig.ioinfo.make(str(cpath))
        cnf = self.psr.load(ioi, ac_ignore_missing=True)
        self.assertEqual(cnf, MZERO)
        self.assertTrue(isinstance(cnf, type(MZERO)))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import collections

from ... import base
from . import common


class MyDict(collections.OrderedDict):
    pass


class TestCase(common.TestCase):

    def test_multi_load_from_empty_path_list(self):
        self.assertEqual(self.target_fn([]), base.NULL_CNTNR)

    def test_multi_load_from_glob_path_str(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn((str(i) for i in tdata.inputs), **tdata.opts),
                tdata.exp
            )

    def test_multi_load_from_streams(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn((i.open() for i in tdata.inputs), **tdata.opts),
                tdata.exp
            )

    def test_multi_load_to_ac_dict(self):
        for tdata in self.each_data():
            res = self.target_fn(tdata.inputs, ac_dict=MyDict, **tdata.opts)
            self.assertEqual(res, tdata.exp, tdata)
            self.assertTrue(isinstance(res, MyDict))

    def test_multi_load_with_wrong_merge_strategy(self):
        for tdata in self.each_data():
            with self.assertRaises(ValueError):
                self.target_fn(tdata.inputs, ac_merge='wrong_merge_strategy')

    def test_multi_load_with_ignore_missing_option(self):
        paths = [
            'file_not_exist_0.json',
            'file_not_exist_1.json',
            'file_not_exist_2.json',
        ]
        with self.assertRaises(FileNotFoundError):
            self.target_fn(paths)

        self.assertEqual(
            self.target_fn(paths, ac_ignore_missing=True),
            base.NULL_CNTNR
        )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Utility functions in anyconfig.backend.base.
"""
import functools
import pathlib
import typing


def not_implemented(*_args, **_kwargs) -> None:
    """
    Utility function to raise NotImplementedError.
    """
    raise NotImplementedError()


def ensure_outdir_exists(filepath: typing.Union[str, pathlib.Path]) -> None:
    """
    Make dir to dump 'filepath' if that dir does not exist.

    :param filepath: path of file to dump
    """
    pathlib.Path(filepath).parent.mkdir(parents=True, exist_ok=True)


def to_method(func: typing.Callable[..., typing.Any]
              ) -> typing.Callable[..., typing.Any]:
    """
    Lift :func:`func` to a method; it will be called with the first argument
    'self' ignored.

    :param func: Any callable object
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """Wrapper function.
        """
        return func(*args[1:], **kwargs)

    return wrapper

# vim:sw=4:ts=4:et:
